from __future__ import annotations

from pandas import DataFrame

from metricflow.filters.time_constraint import TimeRangeConstraint
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.protocols.sql_client import SqlClient
from metricflow.time.time_constants import ISO8601_PYTHON_TS_FORMAT


def test_date_spine_date_range(  # noqa: D
    sql_client: SqlClient,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    create_source_tables: None,
) -> None:
    time_spine_source = simple_semantic_manifest_lookup.time_spine_source
    range_df: DataFrame = sql_client.query(
        f"""\
        SELECT
            MIN({time_spine_source.time_column_name})
            , MAX({time_spine_source.time_column_name})
        FROM {time_spine_source.spine_table.sql}
        """,
    )
    assert range_df.shape == (1, 2), f"Expected 1 row with 2 columns in range dataframe, got {range_df}"
    date_range = tuple(range_df.squeeze())

    assert tuple(map(lambda x: x.strftime(ISO8601_PYTHON_TS_FORMAT), date_range)) == (
        TimeRangeConstraint.ALL_TIME_BEGIN().strftime(ISO8601_PYTHON_TS_FORMAT),
        TimeRangeConstraint.ALL_TIME_END().strftime(ISO8601_PYTHON_TS_FORMAT),
    )



from __future__ import annotations

import logging

import pytest
from _pytest.fixtures import FixtureRequest
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.references import EntityReference, MeasureReference, MetricReference

from metricflow.model.semantics.linkable_element_properties import LinkableElementProperties
from metricflow.model.semantics.metric_lookup import MetricLookup
from metricflow.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.snapshot_utils import assert_linkable_element_set_snapshot_equal, assert_object_snapshot_equal

logger = logging.getLogger(__name__)


@pytest.fixture
def semantic_model_lookup(simple_semantic_manifest: SemanticManifest) -> SemanticModelLookup:  # Noqa: D
    return SemanticModelLookup(
        model=simple_semantic_manifest,
    )


@pytest.fixture
def metric_lookup(  # Noqa: D
    simple_semantic_manifest: SemanticManifest, semantic_model_lookup: SemanticModelLookup
) -> MetricLookup:
    return MetricLookup(semantic_manifest=simple_semantic_manifest, semantic_model_lookup=semantic_model_lookup)


def test_get_names(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    semantic_model_lookup: SemanticModelLookup,
) -> None:
    assert_object_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        obj_id="result0",
        obj={
            "dimension_references": sorted([d.element_name for d in semantic_model_lookup.get_dimension_references()]),
            "measure_references": sorted([m.element_name for m in semantic_model_lookup.measure_references]),
            "entity_references": sorted([i.element_name for i in semantic_model_lookup.get_entity_references()]),
        },
    )


def test_get_elements(semantic_model_lookup: SemanticModelLookup) -> None:  # noqa: D
    for dimension_reference in semantic_model_lookup.get_dimension_references():
        assert (
            semantic_model_lookup.get_dimension(dimension_reference=dimension_reference).reference
            == dimension_reference
        )
    for measure_reference in semantic_model_lookup.measure_references:
        measure_reference = MeasureReference(element_name=measure_reference.element_name)
        assert semantic_model_lookup.get_measure(measure_reference=measure_reference).reference == measure_reference


def test_get_semantic_models_for_measure(semantic_model_lookup: SemanticModelLookup) -> None:  # noqa: D
    bookings_sources = semantic_model_lookup.get_semantic_models_for_measure(MeasureReference(element_name="bookings"))
    assert len(bookings_sources) == 1
    assert bookings_sources[0].name == "bookings_source"

    views_sources = semantic_model_lookup.get_semantic_models_for_measure(MeasureReference(element_name="views"))
    assert len(views_sources) == 1
    assert views_sources[0].name == "views_source"

    listings_sources = semantic_model_lookup.get_semantic_models_for_measure(MeasureReference(element_name="listings"))
    assert len(listings_sources) == 1
    assert listings_sources[0].name == "listings_latest"


def test_elements_for_metric(  # noqa: D
    request: FixtureRequest, mf_test_session_state: MetricFlowTestSessionState, metric_lookup: MetricLookup
) -> None:
    assert_object_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        obj_id="result0",
        obj=tuple(
            spec.qualified_name
            for spec in metric_lookup.element_specs_for_metrics(
                [MetricReference(element_name="views")],
                without_any_property=frozenset(
                    {
                        LinkableElementProperties.DERIVED_TIME_GRANULARITY,
                        LinkableElementProperties.METRIC_TIME,
                    }
                ),
            )
        ),
    )


def test_local_linked_elements_for_metric(  # noqa: D
    request: FixtureRequest, mf_test_session_state: MetricFlowTestSessionState, metric_lookup: MetricLookup
) -> None:
    assert_object_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        obj_id="result0",
        obj=tuple(
            spec.qualified_name
            for spec in metric_lookup.element_specs_for_metrics(
                [MetricReference(element_name="listings")],
                with_any_property=frozenset({LinkableElementProperties.LOCAL_LINKED}),
                without_any_property=frozenset({LinkableElementProperties.DERIVED_TIME_GRANULARITY}),
            )
        ),
    )


def test_get_semantic_models_for_entity(semantic_model_lookup: SemanticModelLookup) -> None:  # noqa: D
    entity_reference = EntityReference(element_name="user")
    linked_semantic_models = semantic_model_lookup.get_semantic_models_for_entity(entity_reference=entity_reference)
    assert len(linked_semantic_models) == 8


def test_linkable_set(  # noqa: D
    request: FixtureRequest, mf_test_session_state: MetricFlowTestSessionState, metric_lookup: MetricLookup
) -> None:
    assert_linkable_element_set_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        set_id="result0",
        linkable_element_set=metric_lookup.linkable_set_for_metrics(
            (MetricReference(element_name="views"),),
            without_any_property=frozenset(
                {
                    LinkableElementProperties.DERIVED_TIME_GRANULARITY,
                    LinkableElementProperties.METRIC_TIME,
                }
            ),
        ),
    )


def test_linkable_set_for_common_dimensions_in_different_models(
    request: FixtureRequest, mf_test_session_state: MetricFlowTestSessionState, metric_lookup: MetricLookup
) -> None:
    """Tests case where a metric has dimensions with the same path.

    In this example, "ds" is defined in both "bookings_source" and "views_source".
    """
    assert_linkable_element_set_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        set_id="result0",
        linkable_element_set=metric_lookup.linkable_set_for_metrics(
            (MetricReference(element_name="bookings_per_view"),),
            without_any_property=frozenset(
                {
                    LinkableElementProperties.DERIVED_TIME_GRANULARITY,
                }
            ),
        ),
    )



from __future__ import annotations

import logging
from typing import Optional

import jinja2
from dbt_semantic_interfaces.protocols import WhereFilterIntersection
from dbt_semantic_interfaces.protocols.where_filter import WhereFilter

from metricflow.filters.merge_where import merge_to_single_where_filter
from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.specs.specs import LinkableSpecSet, WhereFilterSpec
from metricflow.specs.where_filter_dimension import WhereFilterDimensionFactory
from metricflow.specs.where_filter_entity import WhereFilterEntityFactory
from metricflow.specs.where_filter_time_dimension import WhereFilterTimeDimensionFactory
from metricflow.sql.sql_bind_parameters import SqlBindParameters

logger = logging.getLogger(__name__)


class RenderSqlTemplateException(Exception):  # noqa: D
    pass


class WhereSpecFactory:
    """Renders the SQL template in the WhereFilter and converts it to a WhereFilterSpec."""

    def __init__(  # noqa: D
        self,
        column_association_resolver: ColumnAssociationResolver,
        bind_parameters: SqlBindParameters = SqlBindParameters(),
    ) -> None:
        self._column_association_resolver = column_association_resolver
        self._bind_parameters = bind_parameters

    def create_from_where_filter_intersection(  # noqa: D
        self, where_filter_intersection: Optional[WhereFilterIntersection]
    ) -> Optional[WhereFilterSpec]:
        if where_filter_intersection is None:
            return None

        where_filter = merge_to_single_where_filter(where_filter_intersection)
        if where_filter is None:
            return None

        return self.create_from_where_filter(where_filter)

    def create_from_where_filter(self, where_filter: WhereFilter) -> WhereFilterSpec:  # noqa: D
        # Used to check that call parameter sets generated by DSI match those generated below.
        call_parameter_sets = where_filter.call_parameter_sets

        dimension_factory = WhereFilterDimensionFactory(call_parameter_sets, self._column_association_resolver)
        time_dimension_factory = WhereFilterTimeDimensionFactory(call_parameter_sets, self._column_association_resolver)
        entity_factory = WhereFilterEntityFactory(call_parameter_sets, self._column_association_resolver)
        try:
            rendered_sql_template = jinja2.Template(
                where_filter.where_sql_template, undefined=jinja2.StrictUndefined
            ).render(
                {
                    "Dimension": dimension_factory.create,
                    "TimeDimension": time_dimension_factory.create,
                    "Entity": entity_factory.create,
                }
            )
        except (jinja2.exceptions.UndefinedError, jinja2.exceptions.TemplateSyntaxError) as e:
            raise RenderSqlTemplateException(
                f"Error while rendering Jinja template:\n{where_filter.where_sql_template}"
            ) from e

        """
        Dimensions that are created with a grain parameter, Dimension(...).grain(...), are
        added to dimension_specs otherwise they are add to time_dimension_factory.time_dimension_specs
        """
        dimension_specs = []
        for dimension in dimension_factory.created:
            if dimension.time_dimension_spec:
                time_dimension_factory.time_dimension_specs.append(dimension.time_dimension_spec)
            else:
                dimension_specs.append(dimension.dimension_spec)

        return WhereFilterSpec(
            where_sql=rendered_sql_template,
            bind_parameters=self._bind_parameters,
            linkable_spec_set=LinkableSpecSet(
                dimension_specs=tuple(dimension_specs),
                time_dimension_specs=tuple(time_dimension_factory.time_dimension_specs),
                entity_specs=tuple(entity_factory.entity_specs),
            ),
        )



from __future__ import annotations

from typing import List, Optional, Sequence

from dbt_semantic_interfaces.call_parameter_sets import FilterCallParameterSets
from dbt_semantic_interfaces.protocols.protocol_hint import ProtocolHint
from dbt_semantic_interfaces.protocols.query_interface import (
    QueryInterfaceTimeDimension,
    QueryInterfaceTimeDimensionFactory,
)
from dbt_semantic_interfaces.type_enums import TimeGranularity
from typing_extensions import override

from metricflow.errors.errors import InvalidQuerySyntax
from metricflow.specs.column_assoc import ColumnAssociationResolver
from metricflow.specs.dimension_spec_resolver import DimensionSpecResolver
from metricflow.specs.specs import TimeDimensionSpec


class WhereFilterTimeDimension(ProtocolHint[QueryInterfaceTimeDimension]):
    """A time dimension that is passed in through the where filter parameter."""

    @override
    def _implements_protocol(self) -> QueryInterfaceTimeDimension:
        return self

    def __init__(self, column_name: str):  # noqa
        self.column_name = column_name

    def __str__(self) -> str:
        """Returns the column name.

        Important in the Jinja sandbox.
        """
        return self.column_name


class WhereFilterTimeDimensionFactory(ProtocolHint[QueryInterfaceTimeDimensionFactory]):
    """Creates a WhereFilterTimeDimension.

    Each call to `create` adds a TimeDimensionSpec to time_dimension_specs.
    """

    @override
    def _implements_protocol(self) -> QueryInterfaceTimeDimensionFactory:
        return self

    def __init__(  # noqa
        self,
        call_parameter_sets: FilterCallParameterSets,
        column_association_resolver: ColumnAssociationResolver,
    ):
        self._call_parameter_sets = call_parameter_sets
        self._column_association_resolver = column_association_resolver
        self._dimension_spec_resolver = DimensionSpecResolver(call_parameter_sets)
        self.time_dimension_specs: List[TimeDimensionSpec] = []

    def create(
        self,
        time_dimension_name: str,
        time_granularity_name: str,
        entity_path: Sequence[str] = (),
        descending: Optional[bool] = None,
        date_part_name: Optional[str] = None,
    ) -> WhereFilterTimeDimension:
        """Create a WhereFilterTimeDimension."""
        if descending:
            raise InvalidQuerySyntax(
                "Can't set descending in the where clause. Try setting descending in the order_by clause instead"
            )
        if date_part_name:
            raise InvalidQuerySyntax("date_part_name isn't currently supported in the where parameter")
        time_dimension_spec = self._dimension_spec_resolver.resolve_time_dimension_spec(
            time_dimension_name, TimeGranularity(time_granularity_name), entity_path
        )
        self.time_dimension_specs.append(time_dimension_spec)
        column_name = self._column_association_resolver.resolve_spec(time_dimension_spec).column_name
        return WhereFilterTimeDimension(column_name)



from __future__ import annotations

import datetime as dt
import logging
from typing import Dict, List, Optional

from dateutil.parser import parse
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.validations.semantic_manifest_validator import SemanticManifestValidator
from dbt_semantic_interfaces.validations.validator_helpers import SemanticManifestValidationResults

from metricflow.engine.metricflow_engine import (
    MetricFlowEngine,
    MetricFlowExplainResult,
    MetricFlowQueryRequest,
    MetricFlowQueryResult,
)
from metricflow.engine.models import Dimension, Metric
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.optimizer.optimization_levels import SqlQueryOptimizationLevel

logger = logging.getLogger(__name__)


class MetricFlowClient:
    """MetricFlow Python client for running basic queries and other standard commands."""

    def __init__(
        self,
        sql_client: SqlClient,
        semantic_manifest: SemanticManifest,
    ):
        """Initializer for MetricFlowClient.

        Args:
            sql_client: Client that is connected to your data warehouse.
            semantic_manifest: Model containing all the information about your metric configs.
        """
        self.sql_client = sql_client
        self.semantic_manifest = semantic_manifest
        self.semantic_manifest_lookup = SemanticManifestLookup(self.semantic_manifest)
        self.engine = MetricFlowEngine(
            semantic_manifest_lookup=self.semantic_manifest_lookup,
            sql_client=self.sql_client,
        )

    def _create_mf_request(
        self,
        metrics: List[str],
        dimensions: List[str] = [],
        limit: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        where: Optional[str] = None,
        order: Optional[List[str]] = None,
        as_table: Optional[str] = None,
        sql_optimization_level: int = 4,
    ) -> MetricFlowQueryRequest:
        """Build MetricFlowQueryRequest given common query parameters."""
        parsed_optimization_level = SqlQueryOptimizationLevel(f"O{sql_optimization_level}")
        parsed_start_time = _convert_to_datetime(start_time)
        parsed_end_time = _convert_to_datetime(end_time)
        return MetricFlowQueryRequest.create_with_random_request_id(
            metric_names=metrics,
            group_by_names=dimensions,
            limit=limit,
            time_constraint_start=parsed_start_time,
            time_constraint_end=parsed_end_time,
            where_constraint=where,
            order_by_names=order,
            output_table=as_table,
            sql_optimization_level=parsed_optimization_level,
        )

    def query(
        self,
        metrics: List[str],
        dimensions: List[str] = [],
        limit: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        where: Optional[str] = None,
        order: Optional[List[str]] = None,
        as_table: Optional[str] = None,
        sql_optimization_level: int = 4,
    ) -> MetricFlowQueryResult:
        """Makes a query for a metric.

        Args:
            metrics: Names of the metrics to query.
            dimensions: Names of the dimensions and entities to query.
            limit: Limit the result to this many rows.
            start_time: Get data for the start of this time range.
            end_time: Get data for the end of this time range.
            where: A SQL string using group by names that can be used like a where clause on the output data.
            order: metric and group by names to order by. A "-" can be used to specify reverse order e.g. "-ds"
            as_table: If specified, output the result data to this table instead of a result dataframe.
            sql_optimization_level: The level of optimization for the generated SQL. Pass integer from 0-4.

        Returns:
            MetricFlowQueryResult that contains the result and context of the query.
        """
        mf_request = self._create_mf_request(
            metrics=metrics,
            dimensions=dimensions,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
            where=where,
            order=order,
            as_table=as_table,
            sql_optimization_level=sql_optimization_level,
        )
        return self.engine.query(mf_request=mf_request)

    def explain(
        self,
        metrics: List[str],
        dimensions: List[str] = [],
        limit: Optional[int] = None,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
        where: Optional[str] = None,
        order: Optional[List[str]] = None,
        as_table: Optional[str] = None,
        sql_optimization_level: int = 4,
    ) -> MetricFlowExplainResult:
        """Returns the plan for resolving a query.

        Args:
            metrics: Names of the metrics to query.
            dimensions: Names of the dimensions and entities to query.
            limit: Limit the result to this many rows.
            start_time: Get data for the start of this time range.
            end_time: Get data for the end of this time range.
            where: A SQL string using group by names that can be used like a where clause on the output data.
            order: metric and group by names to order by. A "-" can be used to specify reverse order e.g. "-ds"
            as_table: If specified, output the result data to this table instead of a result dataframe.
            sql_optimization_level: The level of optimization for the generated SQL. Pass integer from 0-4.

        Returns:
            MetricFlowExplainResult that contains the context of the query.
        """
        mf_request = self._create_mf_request(
            metrics=metrics,
            dimensions=dimensions,
            limit=limit,
            start_time=start_time,
            end_time=end_time,
            where=where,
            order=order,
            as_table=as_table,
            sql_optimization_level=sql_optimization_level,
        )
        return self.engine.explain(mf_request=mf_request)

    def list_metrics(self) -> Dict[str, Metric]:
        """Retrieves a list of metric names.

        Returns:
            A dictionary with metric names as the key and the corresponding Metric object as the value.
        """
        return {m.name: m for m in self.engine.list_metrics()}

    def list_dimensions(self, metric_names: List[str]) -> List[Dimension]:
        """Retrieves a list of all common dimensions for metric_names.

        "simple" dimensions are the ones that people expect from a UI perspective. For example, if "ds" is a time
        dimension at a day granularity, this would not list "ds__week".

        Args:
            metric_names: Names of metrics to get common dimensions from.

        Returns:
            A list of Dimension objects containing metadata.
        """
        return self.engine.simple_dimensions_for_metrics(metric_names=metric_names)

    def get_dimension_values(
        self,
        metric_names: List[str],
        dimension_name: str,
        start_time: Optional[str] = None,
        end_time: Optional[str] = None,
    ) -> List[str]:
        """Retrieves a list of dimension values given a [metric_name, dimension_name].

        Args:
            metric_names: Names of metrics that contain the group_by.
            dimension_name: Name of group_by to get values from.
            start_time: Get data for the start of this time range.
            end_time: Get data for the end of this time range.

        Returns:
            A list of dimension values as string.
        """
        parsed_start_time = _convert_to_datetime(start_time)
        parsed_end_time = _convert_to_datetime(end_time)
        return self.engine.get_dimension_values(
            metric_names=metric_names,
            get_group_by_values=dimension_name,
            time_constraint_start=parsed_start_time,
            time_constraint_end=parsed_end_time,
        )

    def validate_configs(self) -> SemanticManifestValidationResults:
        """Validate a model according to configured rules.

        Returns:
            Tuple of validation issues with the model provided.
        """
        return SemanticManifestValidator[SemanticManifest]().validate_semantic_manifest(self.semantic_manifest)


def _convert_to_datetime(datetime_str: Optional[str]) -> Optional[dt.datetime]:
    """Callback to convert string to datetime given as an iso8601 timestamp."""
    if datetime_str is None:
        return None

    try:
        return parse(datetime_str)
    except Exception:
        raise ValueError(f"'{datetime_str}' is not a valid iso8601 timestamp")



from __future__ import annotations

import logging

from _pytest.fixtures import FixtureRequest
from dbt_semantic_interfaces.references import SemanticModelReference

from metricflow.protocols.sql_client import SqlClient
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.snapshot_utils import assert_spec_set_snapshot_equal
from metricflow.test.sql.compare_sql_plan import assert_rendered_sql_equal

logger = logging.getLogger(__name__)


def test_convert_table_semantic_model_without_measures(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    """Simple test for converting a table semantic model. Since there are no measures, primary time is not checked."""
    users_data_set = consistent_id_object_repository.simple_model_data_sets["users_latest"]

    assert_spec_set_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        set_id="result0",
        spec_set=users_data_set.instance_set.spec_set,
    )
    assert users_data_set.semantic_model_reference == SemanticModelReference(semantic_model_name="users_latest")
    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan_id="plan0",
        select_node=users_data_set.sql_select_node,
        sql_client=sql_client,
    )


def test_convert_table_semantic_model_with_measures(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    """Complete test of table semantic model conversion. This includes the full set of measures/entities/dimensions.

    Measures trigger a primary time dimension validation. Additionally, this includes both categorical and time
    dimension types, which should cover most, if not all, of the table source branches in the target class.
    """
    id_verifications_data_set = consistent_id_object_repository.simple_model_data_sets["id_verifications"]

    assert_spec_set_snapshot_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        set_id="result0",
        spec_set=id_verifications_data_set.instance_set.spec_set,
    )

    assert id_verifications_data_set.semantic_model_reference == SemanticModelReference(
        semantic_model_name="id_verifications"
    )
    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan_id="plan0",
        select_node=id_verifications_data_set.sql_select_node,
        sql_client=sql_client,
    )


def test_convert_query_semantic_model(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    sql_client: SqlClient,
    default_sql_plan_renderer: SqlQueryPlanRenderer,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    bookings_data_set = consistent_id_object_repository.simple_model_data_sets["revenue"]

    assert_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan_id="plan0",
        select_node=bookings_data_set.sql_select_node,
        sql_client=sql_client,
    )



from __future__ import annotations

import logging
import textwrap
from collections import namedtuple

import pytest
from dbt_semantic_interfaces.parsing.objects import YamlConfigFile
from dbt_semantic_interfaces.references import EntityReference
from dbt_semantic_interfaces.test_utils import as_datetime
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.errors.errors import UnableToSatisfyQueryError
from metricflow.filters.time_constraint import TimeRangeConstraint
from metricflow.query.query_exceptions import InvalidQueryException
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.specs.query_param_implementations import (
    DimensionOrEntityParameter,
    MetricParameter,
    OrderByParameter,
    TimeDimensionParameter,
)
from metricflow.specs.specs import (
    DimensionSpec,
    EntitySpec,
    MetricSpec,
    OrderBySpec,
    TimeDimensionSpec,
)
from metricflow.test.fixtures.model_fixtures import query_parser_from_yaml
from metricflow.test.model.example_project_configuration import EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE
from metricflow.test.time.metric_time_dimension import MTD
from metricflow.time.date_part import DatePart
from metricflow.time.time_granularity_solver import RequestTimeGranularityException

logger = logging.getLogger(__name__)


BOOKINGS_YAML = textwrap.dedent(
    """\
    semantic_model:
      name: bookings_source

      node_relation:
        schema_name: some_schema
        alias: bookings_source_table

      defaults:
        agg_time_dimension: ds

      measures:
        - name: bookings
          expr: "1"
          agg: sum
          create_metric: true

      dimensions:
        - name: is_instant
          type: categorical
        - name: ds
          type: time
          type_params:
            time_granularity: day

      primary_entity: booking

      entities:
        - name: listing
          type: foreign
          expr: listing_id
    """
)


REVENUE_YAML = textwrap.dedent(
    """\
    semantic_model:
      name: revenue_source
      description: revenue

      node_relation:
        schema_name: some_schema
        alias: fct_revenue_table

      defaults:
        agg_time_dimension: ds

      measures:
        - name: revenue
          expr: revenue
          agg: sum
          create_metric: true

      dimensions:
        - name: ds
          type: time
          expr: created_at
          type_params:
            time_granularity: month
        - name: country
          type: categorical
          expr: country

      primary_entity: company

      entities:
        - name: user
          type: foreign
          expr: user_id
    """
)

METRICS_YAML = textwrap.dedent(
    """\
    ---
    metric:
      name: revenue_cumulative
      description: Cumulative metric for revenue for testing purposes
      type: cumulative
      type_params:
        measure: revenue
        window: 7 days
    ---
    metric:
      name: revenue_sub_10
      description: Derived cumulative metric for revenue for testing purposes
      type: derived
      type_params:
        expr: revenue_cumulative - 10
        metrics:
          - name: revenue_cumulative
    ---
    metric:
      name: revenue_growth_2_weeks
      description: Percentage growth of revenue compared to revenue 2 weeks prior
      type: derived
      type_params:
        expr: (revenue - revenue_2_weeks_ago) / revenue_2_weeks_ago
        metrics:
          - name: revenue
          - name: revenue
            offset_window: 14 days
            alias: revenue_2_weeks_ago
    ---
    metric:
      name: revenue_since_start_of_year
      description: Revenue since start of year
      type: derived
      type_params:
        expr: revenue - revenue_start_of_year
        metrics:
          - name: revenue
          - name: revenue
            offset_to_grain: year
            alias: revenue_start_of_year
    """
)


@pytest.fixture
def bookings_query_parser() -> MetricFlowQueryParser:  # noqa
    bookings_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=BOOKINGS_YAML)
    return query_parser_from_yaml([EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, bookings_yaml_file])


def test_query_parser(bookings_query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    query_spec = bookings_query_parser.parse_and_validate_query(
        metric_names=["bookings"],
        group_by_names=["booking__is_instant", "listing", MTD],
        order_by_names=[MTD, "-bookings"],
    )

    assert query_spec.metric_specs == (MetricSpec(element_name="bookings"),)
    assert query_spec.dimension_specs == (
        DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),
    )
    assert query_spec.time_dimension_specs == (
        TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
    )
    assert query_spec.entity_specs == (EntitySpec(element_name="listing", entity_links=()),)
    assert query_spec.order_by_specs == (
        OrderBySpec(
            instance_spec=TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
            descending=False,
        ),
        OrderBySpec(
            instance_spec=MetricSpec(element_name="bookings"),
            descending=True,
        ),
    )


def test_query_parser_case_insensitivity(bookings_query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    # String params
    query_spec = bookings_query_parser.parse_and_validate_query(
        metric_names=["BOOKINGS"],
        group_by_names=["BOOKING__IS_INSTANT", "LISTING", MTD.upper()],
        order_by_names=[MTD.upper(), "-BOOKINGS"],
    )

    assert query_spec.metric_specs == (MetricSpec(element_name="bookings"),)
    assert query_spec.dimension_specs == (
        DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),
    )
    assert query_spec.time_dimension_specs == (
        TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
    )
    assert query_spec.entity_specs == (EntitySpec(element_name="listing", entity_links=()),)
    assert query_spec.order_by_specs == (
        OrderBySpec(
            instance_spec=TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
            descending=False,
        ),
        OrderBySpec(
            instance_spec=MetricSpec(element_name="bookings"),
            descending=True,
        ),
    )

    # Object params
    Metric = namedtuple("Metric", ["name", "descending"])
    metric = Metric("BOOKINGS", False)
    group_by = (
        DimensionOrEntityParameter("BOOKING__IS_INSTANT"),
        DimensionOrEntityParameter("LISTING"),
        TimeDimensionParameter(MTD.upper()),
    )
    order_by = (
        OrderByParameter(order_by=TimeDimensionParameter(MTD.upper())),
        OrderByParameter(order_by=MetricParameter("BOOKINGS"), descending=True),
    )
    query_spec = bookings_query_parser.parse_and_validate_query(metrics=[metric], group_by=group_by, order_by=order_by)
    assert query_spec.metric_specs == (MetricSpec(element_name="bookings"),)
    assert query_spec.dimension_specs == (
        DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),
    )
    assert query_spec.time_dimension_specs == (
        TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
    )
    assert query_spec.entity_specs == (EntitySpec(element_name="listing", entity_links=()),)
    assert query_spec.order_by_specs == (
        OrderBySpec(
            instance_spec=TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
            descending=False,
        ),
        OrderBySpec(
            instance_spec=MetricSpec(element_name="bookings"),
            descending=True,
        ),
    )


def test_query_parser_invalid_group_by(bookings_query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    with pytest.raises(UnableToSatisfyQueryError):
        bookings_query_parser.parse_and_validate_query(group_by_names=["random_stuff"])


def test_query_parser_with_object_params(bookings_query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    Metric = namedtuple("Metric", ["name", "descending"])
    metric = Metric("bookings", False)
    group_by = (
        DimensionOrEntityParameter("booking__is_instant"),
        DimensionOrEntityParameter("listing"),
        TimeDimensionParameter(MTD),
    )
    order_by = (
        OrderByParameter(order_by=TimeDimensionParameter(MTD)),
        OrderByParameter(order_by=MetricParameter("bookings"), descending=True),
    )
    query_spec = bookings_query_parser.parse_and_validate_query(metrics=[metric], group_by=group_by, order_by=order_by)
    assert query_spec.metric_specs == (MetricSpec(element_name="bookings"),)
    assert query_spec.dimension_specs == (
        DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),
    )
    assert query_spec.time_dimension_specs == (
        TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
    )
    assert query_spec.entity_specs == (EntitySpec(element_name="listing", entity_links=()),)
    assert query_spec.order_by_specs == (
        OrderBySpec(
            instance_spec=TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
            descending=False,
        ),
        OrderBySpec(
            instance_spec=MetricSpec(element_name="bookings"),
            descending=True,
        ),
    )


def test_order_by_granularity_conversion() -> None:
    """Test that the granularity of the primary time dimension in the order by is returned appropriately.

    In the case where the primary time dimension is specified in the order by without a granularity suffix, the order
    by spec returned by the parser should have a granularity appropriate for the queried metrics.
    """
    # "bookings" has a granularity of DAY, "revenue" has a granularity of MONTH
    bookings_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=BOOKINGS_YAML)
    revenue_yaml_file = YamlConfigFile(filepath="inline_for_test_2", contents=REVENUE_YAML)

    query_parser = query_parser_from_yaml(
        [EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, bookings_yaml_file, revenue_yaml_file]
    )
    query_spec = query_parser.parse_and_validate_query(
        metric_names=["bookings", "revenue"], group_by_names=[MTD], order_by_names=[f"-{MTD}"]
    )

    # The lowest common granularity is MONTH, so we expect the PTD in the order by to have that granularity.
    assert (
        OrderBySpec(
            instance_spec=TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.MONTH),
            descending=True,
        ),
    ) == query_spec.order_by_specs


def test_order_by_granularity_no_conversion(bookings_query_parser: MetricFlowQueryParser) -> None:  # noqa: D
    query_spec = bookings_query_parser.parse_and_validate_query(
        metric_names=["bookings"], group_by_names=[MTD], order_by_names=[MTD]
    )

    # The only granularity is DAY, so we expect the PTD in the order by to have that granularity.
    assert (
        OrderBySpec(
            instance_spec=TimeDimensionSpec(element_name=MTD, entity_links=(), time_granularity=TimeGranularity.DAY),
            descending=False,
        ),
    ) == query_spec.order_by_specs


def test_time_range_constraint_conversion() -> None:
    """Test that the returned time constraint in the query spec is adjusted to match the granularity of the query."""
    bookings_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=BOOKINGS_YAML)
    revenue_yaml_file = YamlConfigFile(filepath="inline_for_test_2", contents=REVENUE_YAML)

    query_parser = query_parser_from_yaml(
        [EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, bookings_yaml_file, revenue_yaml_file]
    )

    # "bookings" has a granularity of DAY, "revenue" has a granularity of MONTH
    query_spec = query_parser.parse_and_validate_query(
        metric_names=["bookings", "revenue"],
        group_by_names=[MTD],
        time_constraint_start=as_datetime("2020-01-15"),
        time_constraint_end=as_datetime("2020-02-15"),
    )

    assert (
        TimeRangeConstraint(start_time=as_datetime("2020-01-01"), end_time=as_datetime("2020-02-29"))
    ) == query_spec.time_range_constraint


def test_parse_and_validate_where_constraint_dims(bookings_query_parser: MetricFlowQueryParser) -> None:
    """Test that the returned time constraint in the query spec is adjusted to match the granularity of the query."""
    # check constraint on invalid_dim raises UnableToSatisfyQueryError
    with pytest.raises(UnableToSatisfyQueryError):
        bookings_query_parser.parse_and_validate_query(
            metric_names=["bookings"],
            group_by_names=[MTD],
            time_constraint_start=as_datetime("2020-01-15"),
            time_constraint_end=as_datetime("2020-02-15"),
            where_constraint_str="{{ Dimension('booking__invalid_dim') }} = '1'",
        )

    with pytest.raises(InvalidQueryException):
        bookings_query_parser.parse_and_validate_query(
            metric_names=["bookings"],
            group_by_names=[MTD],
            time_constraint_start=as_datetime("2020-01-15"),
            time_constraint_end=as_datetime("2020-02-15"),
            where_constraint_str="{{ Dimension('invalid_format') }} = '1'",
        )

    query_spec = bookings_query_parser.parse_and_validate_query(
        metric_names=["bookings"],
        group_by_names=[MTD],
        time_constraint_start=as_datetime("2020-01-15"),
        time_constraint_end=as_datetime("2020-02-15"),
        where_constraint_str="{{ Dimension('booking__is_instant') }} = '1'",
    )
    assert (
        DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),))
        not in query_spec.dimension_specs
    )


def test_parse_and_validate_where_constraint_metric_time() -> None:
    """Test that granularity of metric_time reference in where constraint is at least that of the ds dimension."""
    revenue_yaml_file = YamlConfigFile(filepath="inline_for_test_2", contents=REVENUE_YAML)

    query_parser = query_parser_from_yaml([EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, revenue_yaml_file])
    with pytest.raises(RequestTimeGranularityException):
        query_parser.parse_and_validate_query(
            metric_names=["revenue"],
            group_by_names=[MTD],
            where_constraint_str="{{ TimeDimension('metric_time', 'day') }} > '2020-01-15'",
        )


def test_parse_and_validate_metric_constraint_dims() -> None:
    """Test that the returned time constraint in the query spec is adjusted to match the granularity of the query."""
    revenue_yaml_file = YamlConfigFile(filepath="inline_for_test_2", contents=REVENUE_YAML)
    query_parser = query_parser_from_yaml([EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, revenue_yaml_file])

    # check constraint on invalid_dim raises UnableToSatisfyQueryError
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["metric_with_invalid_constraint"],
            group_by_names=[MTD],
            time_constraint_start=as_datetime("2020-01-15"),
            time_constraint_end=as_datetime("2020-02-15"),
        )


def test_derived_metric_query_parsing() -> None:
    """Test derived metric inputs are properly validated."""
    bookings_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=BOOKINGS_YAML)
    metrics_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=METRICS_YAML)
    revenue_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=REVENUE_YAML)
    query_parser = query_parser_from_yaml(
        [EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, bookings_yaml_file, revenue_yaml_file, metrics_yaml_file]
    )
    # Attempt to query with no dimension
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["revenue_sub_10"],
            group_by_names=[],
        )

    # Attempt to query with non-time dimension
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["revenue_sub_10"],
            group_by_names=["country"],
        )

    # Query with time dimension
    query_parser.parse_and_validate_query(
        metric_names=["revenue_sub_10"],
        group_by_names=[MTD],
    )


def test_derived_metric_with_offset_parsing() -> None:
    """Test that querying derived metrics with a time offset requires a time dimension."""
    revenue_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=REVENUE_YAML)
    metrics_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=METRICS_YAML)
    query_parser = query_parser_from_yaml(
        [EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, revenue_yaml_file, metrics_yaml_file]
    )
    # Attempt to query with no dimension
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["revenue_growth_2_weeks"],
            group_by_names=[],
        )

    # Attempt to query with non-time dimension
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["revenue_growth_2_weeks"],
            group_by_names=["country"],
        )

    # Query with time dimension
    query_parser.parse_and_validate_query(
        metric_names=["revenue_growth_2_weeks"],
        group_by_names=[MTD],
    )


def test_date_part_parsing() -> None:
    """Test that querying with a date_part verifies compatibility with time_granularity."""
    revenue_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=REVENUE_YAML)
    metrics_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=METRICS_YAML)
    query_parser = query_parser_from_yaml(
        [EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, revenue_yaml_file, metrics_yaml_file]
    )

    # Date part is incompatible with metric's defined time granularity
    with pytest.raises(RequestTimeGranularityException):
        query_parser.parse_and_validate_query(
            metric_names=["revenue"],
            group_by=(TimeDimensionParameter(name="metric_time", date_part=DatePart.DOW),),
        )

    # Can't query date part for cumulative metrics
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["revenue_cumulative"],
            group_by=(TimeDimensionParameter(name="metric_time", date_part=DatePart.YEAR),),
        )

    # Can't query date part for metrics with offset to grain
    with pytest.raises(UnableToSatisfyQueryError):
        query_parser.parse_and_validate_query(
            metric_names=["revenue_since_start_of_year"],
            group_by=(TimeDimensionParameter(name="metric_time", date_part=DatePart.MONTH),),
        )

    # Requested granularity doesn't match resolved granularity
    with pytest.raises(RequestTimeGranularityException):
        query_parser.parse_and_validate_query(
            metric_names=["revenue"],
            group_by=(
                TimeDimensionParameter(name="metric_time", grain=TimeGranularity.YEAR, date_part=DatePart.MONTH),
            ),
        )

    # Date part is compatible
    query_parser.parse_and_validate_query(
        metric_names=["revenue"],
        group_by=(TimeDimensionParameter(name="metric_time", date_part=DatePart.MONTH),),
    )



from __future__ import annotations

import logging
from typing import Set, Union

import pandas as pd
import pytest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.protocols.sql_client import SqlClient, SqlEngine
from metricflow.random_id import random_id
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.test.compare_df import assert_dataframes_equal
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.fixtures.sql_clients.ddl_sql_client import SqlClientWithDDLMethods

logger = logging.getLogger(__name__)


def _random_table() -> str:
    return f"test_table_{random_id()}"


def _select_x_as_y(x: int = 1, y: str = "y") -> str:  # noqa: D
    return f"SELECT {x} AS {y}"


def _check_1col(df: pd.DataFrame, col: str = "y", vals: Set[Union[int, str]] = {1}) -> None:
    """Helper to check that 1 column has the same value and a case-insensitive matching name.

    We lower-case the names due to snowflake's tendency to capitalize things. This isn't ideal but it'll do for now.
    """
    df.columns = df.columns.str.lower()
    col = col.lower()
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (len(vals), 1)
    assert df.columns.tolist() == [col]
    assert set(df[col]) == vals


def test_query(sql_client: SqlClient) -> None:  # noqa: D
    df = sql_client.query(_select_x_as_y())
    _check_1col(df)


def test_select_one_query(sql_client: SqlClient) -> None:  # noqa: D
    sql_client.query("SELECT 1")
    with pytest.raises(Exception):
        sql_client.query("this is garbage")


def test_create_table_from_dataframe(  # noqa: D
    mf_test_session_state: MetricFlowTestSessionState, ddl_sql_client: SqlClientWithDDLMethods
) -> None:
    expected_df = pd.DataFrame(
        columns=["int_col", "str_col", "float_col", "bool_col", "time_col"],
        data=[
            (1, "abc", 1.23, False, "2020-01-01"),
            (2, "def", 4.56, True, "2020-01-02"),
            (3, "ghi", 1.1, False, None),  # Test NaT type
            (None, "jkl", None, True, "2020-01-03"),  # Test NaN types
            (3, None, 1.2, None, "2020-01-04"),  # Test None types for NA conversions
        ],
    )
    sql_table = SqlTable(schema_name=mf_test_session_state.mf_system_schema, table_name=_random_table())
    ddl_sql_client.create_table_from_dataframe(sql_table=sql_table, df=expected_df)

    actual_df = ddl_sql_client.query(f"SELECT * FROM {sql_table.sql}")
    assert_dataframes_equal(
        actual=actual_df,
        expected=expected_df,
        compare_names_using_lowercase=ddl_sql_client.sql_engine_type is SqlEngine.SNOWFLAKE,
    )


def test_percent_signs_in_query(sql_client: SqlClient) -> None:
    """Note: this only syntax works for Datbricks if no execution params are passed."""
    stmt = "SELECT foo FROM ( SELECT 'abba' AS foo ) source0 WHERE foo LIKE '%a'"
    sql_client.query(stmt)
    df = sql_client.query(stmt)
    _check_1col(df, col="foo", vals={"abba"})


@pytest.fixture()
def example_df() -> pd.DataFrame:
    """Data frame containing data of different types for testing. DateTime would be good to add."""
    return pd.DataFrame(
        columns=["int_col", "str_col", "float_col", "bool_col"],
        data=[
            (1, "abc", 1.23, False),
            (2, "def", 4.56, True),
        ],
    )


def test_dry_run(mf_test_session_state: MetricFlowTestSessionState, sql_client: SqlClient) -> None:  # noqa: D
    test_table = SqlTable(schema_name=mf_test_session_state.mf_system_schema, table_name=_random_table())

    stmt = f"CREATE TABLE {test_table.sql} AS SELECT 1 AS foo"
    sql_client.dry_run(stmt)
    with pytest.raises(expected_exception=Exception) as excinfo:
        sql_client.execute(stmt=f"SELECT * FROM {test_table.sql}")

    exception_message = repr(excinfo.value)
    match = (
        test_table.table_name
        if sql_client.sql_engine_type is not SqlEngine.SNOWFLAKE
        else str.upper(test_table.table_name)
    )
    assert (
        exception_message.find(f"{match}") != -1
    ), f"Expected an exception about table {match} not found, but got `{exception_message}`"


def test_dry_run_of_bad_query_raises_exception(sql_client: SqlClient) -> None:  # noqa: D
    bad_stmt = "SELECT bad_col"
    # Tests that a bad query raises an exception. Different engines may raise different exceptions e.g.
    # ProgrammingError, OperationalError, google.api_core.exceptions.BadRequest, etc.
    with pytest.raises(Exception, match=r"(?i)bad_col"):
        sql_client.dry_run(bad_stmt)


def test_update_params_with_same_item() -> None:  # noqa: D
    bind_params0 = SqlBindParameters.create_from_dict({"key": "value"})
    bind_params1 = SqlBindParameters.create_from_dict({"key": "value"})

    bind_params0.combine(bind_params1)


def test_update_params_with_same_key_different_values() -> None:  # noqa: D
    bind_params0 = SqlBindParameters.create_from_dict(({"key": "value0"}))
    bind_params1 = SqlBindParameters.create_from_dict(({"key": "value1"}))

    with pytest.raises(RuntimeError):
        bind_params0.combine(bind_params1)



from __future__ import annotations

import pathlib
import tempfile
from typing import Generator, Optional, Sequence

import click
import pytest
from click.testing import CliRunner, Result
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.test_utils import as_datetime
from typing_extensions import override

from metricflow.cli.cli_context import CLIContext
from metricflow.cli.dbt_connectors.dbt_config_accessor import dbtArtifacts, dbtProjectMetadata
from metricflow.engine.metricflow_engine import MetricFlowEngine
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.protocols.sql_client import SqlClient
from metricflow.test.time.configurable_time_source import ConfigurableTimeSource


class FakeCLIContext(CLIContext):
    """Fake context for testing. Manually initialize or override params used by CLIContext as appropriate.

    Note - this construct should not exist. It bypasses a fundamental initialization process within the CLI which
    can currently only be tested manually. Eventually, we need to refactor the CLI itself to allow for more robust
    testing. This will, however, allow us to test the invocation of CLI commands assuming the required configuration
    components are all in place.
    """

    def __init__(self) -> None:
        """Initializer configures only the two required parameters and leaves the remainder for inline override."""
        self.verbose = False
        self._mf: Optional[MetricFlowEngine] = None
        self._sql_client: Optional[SqlClient] = None
        self._semantic_manifest: Optional[SemanticManifest] = None
        self._semantic_manifest_lookup: Optional[SemanticManifestLookup] = None
        self._log_file_path: Optional[pathlib.Path] = None

    @property
    @override
    def dbt_artifacts(self) -> dbtArtifacts:
        raise NotImplementedError("FakeCLIContext does not load full dbt artifacts!")

    @property
    @override
    def dbt_project_metadata(self) -> dbtProjectMetadata:
        raise NotImplementedError("FakeCLIContext does not load dbt project metadata!")

    @property
    @override
    def log_file_path(self) -> pathlib.Path:
        assert self._log_file_path
        return self._log_file_path

    @property
    @override
    def sql_client(self) -> SqlClient:
        assert self._sql_client, "Must set _sql_client before use!"
        return self._sql_client

    @property
    @override
    def mf(self) -> MetricFlowEngine:
        assert self._mf, "Must set _mf before use!"
        return self._mf

    @property
    @override
    def semantic_manifest(self) -> SemanticManifest:
        assert self._semantic_manifest, "Must set _semantic_manifest before use!"
        return self._semantic_manifest


@pytest.fixture
def cli_context(  # noqa: D
    sql_client: SqlClient,
    simple_semantic_manifest: SemanticManifest,
    create_source_tables: bool,
) -> Generator[CLIContext, None, None]:
    semantic_manifest_lookup = SemanticManifestLookup(simple_semantic_manifest)
    mf_engine = MetricFlowEngine(
        semantic_manifest_lookup=semantic_manifest_lookup,
        sql_client=sql_client,
        column_association_resolver=DunderColumnAssociationResolver(semantic_manifest_lookup=semantic_manifest_lookup),
        time_source=ConfigurableTimeSource(as_datetime("2020-01-01")),
    )
    context = FakeCLIContext()
    context._mf = mf_engine
    context._sql_client = sql_client
    context._semantic_manifest = simple_semantic_manifest
    context._semantic_manifest_lookup = semantic_manifest_lookup
    with tempfile.NamedTemporaryFile() as file:
        context._log_file_path = pathlib.Path(file.name)
        yield context


class MetricFlowCliRunner(CliRunner):
    """Custom CliRunner class to handle passing context."""

    def __init__(self, cli_context: CLIContext) -> None:  # noqa: D
        self.cli_context = cli_context
        super().__init__()

    def run(self, cli: click.BaseCommand, args: Optional[Sequence[str]] = None) -> Result:  # noqa: D
        # TODO: configure CLI to use a dbt_project fixture
        dummy_dbt_project = pathlib.Path("dbt_project.yml")
        dummy_dbt_project.touch()
        result = super().invoke(cli, args, obj=self.cli_context)
        dummy_dbt_project.unlink()
        return result


@pytest.fixture
def cli_runner(cli_context: CLIContext) -> MetricFlowCliRunner:  # noqa: D
    return MetricFlowCliRunner(cli_context=cli_context)



from __future__ import annotations

import logging
import os
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, List, Optional, Sequence

import pytest
from dbt_semantic_interfaces.implementations.semantic_manifest import PydanticSemanticManifest
from dbt_semantic_interfaces.parsing.dir_to_model import (
    SemanticManifestBuildResult,
    parse_directory_of_yaml_files_to_semantic_manifest,
    parse_yaml_files_to_validation_ready_semantic_manifest,
)
from dbt_semantic_interfaces.parsing.objects import YamlConfigFile
from dbt_semantic_interfaces.protocols.semantic_manifest import SemanticManifest
from dbt_semantic_interfaces.protocols.semantic_model import SemanticModel
from dbt_semantic_interfaces.transformations.pydantic_rule_set import PydanticSemanticManifestTransformRuleSet
from dbt_semantic_interfaces.transformations.semantic_manifest_transformer import PydanticSemanticManifestTransformer
from dbt_semantic_interfaces.validations.semantic_manifest_validator import SemanticManifestValidator

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.source_node import SourceNodeBuilder
from metricflow.dataflow.dataflow_plan import BaseOutput, ReadSqlSourceNode
from metricflow.dataset.convert_semantic_model import SemanticModelToDataSetConverter
from metricflow.dataset.semantic_model_adapter import SemanticModelDataSet
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.query.query_parser import MetricFlowQueryParser
from metricflow.test.fixtures.id_fixtures import IdNumberSpace, patch_id_generators_helper
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState

logger = logging.getLogger(__name__)


def _data_set_to_read_nodes(data_sets: OrderedDict[str, SemanticModelDataSet]) -> OrderedDict[str, ReadSqlSourceNode]:
    """Return a mapping from the name of the semantic model to the dataflow plan node that reads from it."""
    return_dict: OrderedDict[str, ReadSqlSourceNode] = OrderedDict()
    for semantic_model_name, data_set in data_sets.items():
        return_dict[semantic_model_name] = ReadSqlSourceNode(data_set)
        logger.debug(
            f"For semantic model {semantic_model_name}, creating node_id {return_dict[semantic_model_name].node_id}"
        )

    return return_dict


def _data_set_to_source_nodes(
    semantic_manifest_lookup: SemanticManifestLookup, data_sets: OrderedDict[str, SemanticModelDataSet]
) -> Sequence[BaseOutput]:
    source_node_builder = SourceNodeBuilder(semantic_manifest_lookup)
    return source_node_builder.create_from_data_sets(list(data_sets.values()))


def query_parser_from_yaml(yaml_contents: List[YamlConfigFile]) -> MetricFlowQueryParser:
    """Given yaml files, return a query parser using default source nodes, resolvers and time spine source."""
    semantic_manifest_lookup = SemanticManifestLookup(
        parse_yaml_files_to_validation_ready_semantic_manifest(
            yaml_contents, apply_transformations=True
        ).semantic_manifest
    )
    SemanticManifestValidator[SemanticManifest]().checked_validations(semantic_manifest_lookup.semantic_manifest)
    return MetricFlowQueryParser(
        model=semantic_manifest_lookup,
        column_association_resolver=DunderColumnAssociationResolver(semantic_manifest_lookup),
        read_nodes=list(_data_set_to_read_nodes(create_data_sets(semantic_manifest_lookup)).values()),
        node_output_resolver=DataflowPlanNodeOutputDataSetResolver(
            column_association_resolver=DunderColumnAssociationResolver(semantic_manifest_lookup),
            semantic_manifest_lookup=semantic_manifest_lookup,
        ),
    )


@dataclass(frozen=True)
class ConsistentIdObjectRepository:
    """Stores all objects that should have consistent IDs in tests."""

    simple_model_data_sets: OrderedDict[str, SemanticModelDataSet]
    simple_model_read_nodes: OrderedDict[str, ReadSqlSourceNode]
    simple_model_source_nodes: Sequence[BaseOutput]

    multihop_model_read_nodes: OrderedDict[str, ReadSqlSourceNode]
    multihop_model_source_nodes: Sequence[BaseOutput]

    scd_model_data_sets: OrderedDict[str, SemanticModelDataSet]
    scd_model_read_nodes: OrderedDict[str, ReadSqlSourceNode]
    scd_model_source_nodes: Sequence[BaseOutput]

    cyclic_join_read_nodes: OrderedDict[str, ReadSqlSourceNode]
    cyclic_join_source_nodes: Sequence[BaseOutput]

    extended_date_model_read_nodes: OrderedDict[str, ReadSqlSourceNode]
    extended_date_model_source_nodes: Sequence[BaseOutput]


@pytest.fixture(scope="session")
def consistent_id_object_repository(
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
    scd_semantic_manifest_lookup: SemanticManifestLookup,
    cyclic_join_semantic_manifest_lookup: SemanticManifestLookup,
    extended_date_semantic_manifest_lookup: SemanticManifestLookup,
) -> ConsistentIdObjectRepository:  # noqa: D
    """Create objects that have incremental numeric IDs with a consistent value.

    This should use IDs with a high enough value so that when other tests run with ID generators set to 0 at the start
    of the test and create objects, there is no overlap in the IDs.
    """
    with patch_id_generators_helper(start_value=IdNumberSpace.CONSISTENT_ID_REPOSITORY):
        sm_data_sets = create_data_sets(simple_semantic_manifest_lookup)
        multihop_data_sets = create_data_sets(multi_hop_join_semantic_manifest_lookup)
        scd_data_sets = create_data_sets(scd_semantic_manifest_lookup)
        cyclic_join_data_sets = create_data_sets(cyclic_join_semantic_manifest_lookup)
        extended_date_data_sets = create_data_sets(extended_date_semantic_manifest_lookup)

        return ConsistentIdObjectRepository(
            simple_model_data_sets=sm_data_sets,
            simple_model_read_nodes=_data_set_to_read_nodes(sm_data_sets),
            simple_model_source_nodes=_data_set_to_source_nodes(simple_semantic_manifest_lookup, sm_data_sets),
            multihop_model_read_nodes=_data_set_to_read_nodes(multihop_data_sets),
            multihop_model_source_nodes=_data_set_to_source_nodes(
                multi_hop_join_semantic_manifest_lookup, multihop_data_sets
            ),
            scd_model_data_sets=scd_data_sets,
            scd_model_read_nodes=_data_set_to_read_nodes(scd_data_sets),
            scd_model_source_nodes=_data_set_to_source_nodes(
                semantic_manifest_lookup=scd_semantic_manifest_lookup, data_sets=scd_data_sets
            ),
            cyclic_join_read_nodes=_data_set_to_read_nodes(cyclic_join_data_sets),
            cyclic_join_source_nodes=_data_set_to_source_nodes(
                semantic_manifest_lookup=cyclic_join_semantic_manifest_lookup, data_sets=cyclic_join_data_sets
            ),
            extended_date_model_read_nodes=_data_set_to_read_nodes(extended_date_data_sets),
            extended_date_model_source_nodes=_data_set_to_source_nodes(
                semantic_manifest_lookup=extended_date_semantic_manifest_lookup, data_sets=extended_date_data_sets
            ),
        )


def create_data_sets(
    multihop_semantic_manifest_lookup: SemanticManifestLookup,
) -> OrderedDict[str, SemanticModelDataSet]:
    """Convert the SemanticModels in the model to SqlDataSets.

    Key is the name of the semantic model, value is the associated data set.
    """
    # Use ordered dict and sort by name to get consistency when running tests.
    data_sets = OrderedDict()
    semantic_models: Sequence[SemanticModel] = multihop_semantic_manifest_lookup.semantic_manifest.semantic_models
    semantic_models = sorted(semantic_models, key=lambda x: x.name)

    converter = SemanticModelToDataSetConverter(
        column_association_resolver=DunderColumnAssociationResolver(multihop_semantic_manifest_lookup)
    )

    for semantic_model in semantic_models:
        data_sets[semantic_model.name] = converter.create_sql_source_data_set(semantic_model)

    return data_sets


def load_semantic_manifest(
    relative_manifest_path: str,
    template_mapping: Optional[Dict[str, str]] = None,
) -> SemanticManifestBuildResult:
    """Reads the manifest YAMLs from the standard location, applies transformations, runs validations."""
    yaml_file_directory = os.path.join(os.path.dirname(__file__), f"semantic_manifest_yamls/{relative_manifest_path}")
    build_result = parse_directory_of_yaml_files_to_semantic_manifest(
        yaml_file_directory, template_mapping=template_mapping
    )
    validator = SemanticManifestValidator[PydanticSemanticManifest]()
    validator.checked_validations(build_result.semantic_manifest)
    return build_result


@pytest.fixture(scope="session")
def template_mapping(mf_test_session_state: MetricFlowTestSessionState) -> Dict[str, str]:
    """Mapping for template variables in the model YAML files."""
    return {"source_schema": mf_test_session_state.mf_source_schema}


@pytest.fixture(scope="session")
def simple_semantic_manifest_lookup_non_ds(template_mapping: Dict[str, str]) -> SemanticManifestLookup:  # noqa: D
    build_result = load_semantic_manifest("non_sm_manifest", template_mapping)
    return SemanticManifestLookup(build_result.semantic_manifest)


@pytest.fixture(scope="session")
def simple_semantic_manifest_lookup(template_mapping: Dict[str, str]) -> SemanticManifestLookup:  # noqa: D
    build_result = load_semantic_manifest("simple_manifest", template_mapping)
    return SemanticManifestLookup(build_result.semantic_manifest)


@pytest.fixture(scope="session")
def multi_hop_join_semantic_manifest_lookup(template_mapping: Dict[str, str]) -> SemanticManifestLookup:  # noqa: D
    build_result = load_semantic_manifest("multi_hop_join_manifest/partitioned_semantic_models", template_mapping)
    return SemanticManifestLookup(build_result.semantic_manifest)


@pytest.fixture(scope="session")
def unpartitioned_multi_hop_join_semantic_manifest_lookup(  # noqa: D
    template_mapping: Dict[str, str]
) -> SemanticManifestLookup:
    build_result = load_semantic_manifest("multi_hop_join_manifest/unpartitioned_semantic_models", template_mapping)
    return SemanticManifestLookup(build_result.semantic_manifest)


@pytest.fixture(scope="session")
def simple_semantic_manifest(template_mapping: Dict[str, str]) -> SemanticManifest:
    """Model used for many tests."""
    build_result = load_semantic_manifest("simple_manifest", template_mapping)
    return build_result.semantic_manifest


@pytest.fixture(scope="session")
def simple_model__with_primary_transforms(template_mapping: Dict[str, str]) -> SemanticManifest:
    """Model used for tests pre-transformations."""
    build_result = load_semantic_manifest("simple_manifest", template_mapping)
    transformed_model = PydanticSemanticManifestTransformer().transform(
        model=build_result.semantic_manifest,
        ordered_rule_sequences=(PydanticSemanticManifestTransformRuleSet().primary_rules,),
    )
    return transformed_model


@pytest.fixture(scope="session")
def extended_date_semantic_manifest_lookup(template_mapping: Dict[str, str]) -> SemanticManifestLookup:  # noqa: D
    build_result = load_semantic_manifest("extended_date_manifest", template_mapping)
    return SemanticManifestLookup(build_result.semantic_manifest)


@pytest.fixture(scope="session")
def scd_semantic_manifest_lookup(template_mapping: Dict[str, str]) -> SemanticManifestLookup:
    """Initialize semantic model for SCD tests."""
    build_result = load_semantic_manifest("scd_manifest", template_mapping)
    return SemanticManifestLookup(build_result.semantic_manifest)


@pytest.fixture(scope="session")
def data_warehouse_validation_model(template_mapping: Dict[str, str]) -> SemanticManifest:
    """Model used for data warehouse validation tests."""
    build_result = load_semantic_manifest("data_warehouse_validation_manifest", template_mapping)
    return build_result.semantic_manifest


@pytest.fixture(scope="session")
def cyclic_join_semantic_manifest_lookup(template_mapping: Dict[str, str]) -> SemanticManifestLookup:
    """Manifest that contains a potential cycle in the join graph (if not handled properly)."""
    build_result = load_semantic_manifest("cyclic_join_manifest", template_mapping)
    return SemanticManifestLookup(build_result.semantic_manifest)


@pytest.fixture(scope="session")
def node_output_resolver(  # noqa:D
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> DataflowPlanNodeOutputDataSetResolver:
    return DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
    )



# Need to ignore flake8 as there are blank lines in the test outputs, which throws "W293 blank line contains whitespace"
# flake8: noqa

import logging
import textwrap

import pytest
from dbt_semantic_interfaces.parsing.objects import YamlConfigFile

from metricflow.errors.errors import UnableToSatisfyQueryError
from metricflow.test.fixtures.model_fixtures import query_parser_from_yaml
from metricflow.test.model.example_project_configuration import EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE

logger = logging.getLogger(__name__)


EXTENDED_BOOKINGS_YAML = textwrap.dedent(
    """\
    semantic_model:
      name: bookings_source

      node_relation:
        schema_name: some_schema
        alias: bookings_source_table

      defaults:
        agg_time_dimension: ds

      measures:
        - name: bookings
          expr: "1"
          agg: sum
          create_metric: true
        - name: instant_bookings
          expr: is_instant
          agg: sum_boolean
          create_metric: true
        - name: booking_value
          agg: sum
          create_metric: true
        - name: max_booking_value
          agg: max
          expr: booking_value
          create_metric: true
        - name: min_booking_value
          agg: min
          expr: booking_value
          create_metric: true
        - name: bookers
          expr: guest_id
          agg: count_distinct
          create_metric: true
        - name: average_booking_value
          expr: booking_value
          agg: average
          create_metric: true
        - name: booking_payments
          expr: booking_value
          agg: sum
          create_metric: true
        - name: referred_bookings
          expr: referrer_id
          agg: count
          create_metric: true

      dimensions:
        - name: is_instant
          type: categorical
        - name: ds
          type: time
          type_params:
            time_granularity: day

      primary_entity: booking

      entities:
        - name: listing
          type: foreign
          expr: listing_id
    """
)

LISTINGS_YAML = textwrap.dedent(
    """\
    semantic_model:
      name: listings_latest
      description: listings_latest

      node_relation:
        schema_name: schema
        alias: table

      defaults:
        agg_time_dimension: ds

      measures:
        - name: listings
          expr: 1
          agg: sum

      dimensions:
        - name: ds
          type: time
          expr: created_at
          type_params:
            time_granularity: day
        - name: created_at
          type: time
          type_params:
            time_granularity: day
        - name: country_latest
          type: categorical
          expr: country
        - name: capacity_latest
          type: categorical
          expr: capacity

      entities:
        - name: listing
          type: primary
          expr: listing_id
        - name: user
          type: foreign
          expr: user_id
    """
)


def test_nonexistent_metric() -> None:  # noqa: D
    bookings_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=EXTENDED_BOOKINGS_YAML)
    query_parser = query_parser_from_yaml([EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, bookings_yaml_file])

    with pytest.raises(UnableToSatisfyQueryError) as exception_info:
        query_parser.parse_and_validate_query(metric_names=["booking"], group_by_names=["booking__is_instant"])

    assert (
        textwrap.dedent(
            """\
            Unable To Satisfy Query Error: Unknown metric: 'booking'

            Suggestions for 'booking':
                ['bookings',
                 'booking_value',
                 'instant_bookings',
                 'booking_payments',
                 'max_booking_value',
                 'min_booking_value']
            """
        ).rstrip()
        == str(exception_info.value)
    )


def test_non_existent_group_by() -> None:  # noqa: D
    bookings_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=EXTENDED_BOOKINGS_YAML)
    query_parser = query_parser_from_yaml([EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, bookings_yaml_file])

    with pytest.raises(UnableToSatisfyQueryError) as exception_info:
        query_parser.parse_and_validate_query(metric_names=["bookings"], group_by_names=["is_instan"])

    assert (
        textwrap.dedent(
            """\
            Unable To Satisfy Query Error: Unknown element name 'is_instan' in dimension name 'is_instan'

            Suggestions for 'is_instan':
                ['booking__is_instant']
            """
        ).rstrip()
        == str(exception_info.value)
    )


def test_invalid_group_by() -> None:  # noqa: D
    bookings_yaml_file = YamlConfigFile(filepath="inline_for_test_1", contents=EXTENDED_BOOKINGS_YAML)
    listings_yaml_file = YamlConfigFile(filepath="inline_for_test_2", contents=LISTINGS_YAML)

    query_parser = query_parser_from_yaml(
        [EXAMPLE_PROJECT_CONFIGURATION_YAML_CONFIG_FILE, bookings_yaml_file, listings_yaml_file]
    )

    with pytest.raises(UnableToSatisfyQueryError) as exception_info:
        query_parser.parse_and_validate_query(metric_names=["bookings"], group_by_names=["capacity_latest"])

    assert (
        textwrap.dedent(
            """\
            Unable To Satisfy Query Error: Dimensions ['capacity_latest'] cannot be resolved for metrics \
['bookings']. The invalid dimension may not exist, require an ambiguous join (e.g. a join path that can be satisfied \
in multiple ways), or require a fanout join.

            Suggestions for invalid dimension 'capacity_latest':
                ['listing__capacity_latest', 'listing__country_latest']
            """
        ).rstrip()
        == str(exception_info.value)
    )



from __future__ import annotations

import pytest

from metricflow.errors.errors import InvalidQuerySyntax
from metricflow.specs.where_filter_entity import WhereFilterEntity


def test_descending_cannot_be_set() -> None:  # noqa
    with pytest.raises(InvalidQuerySyntax):
        WhereFilterEntity("customer").descending(True)



from __future__ import annotations

import logging
import math
from typing import SupportsFloat

import pandas as pd

logger = logging.getLogger(__name__)


def _dataframes_contain_same_data(
    actual: pd.DataFrame,
    expected: pd.DataFrame,
) -> bool:
    """Compare all elements of the dataframe one by one for equality.

    We use math.isclose() to avoid issues with float comparisons.
    """
    if expected.shape != actual.shape:
        return False

    for c in range(expected.shape[0]):
        for r in range(expected.shape[1]):
            # NaNs can't be compared for equality.
            if pd.isna(expected.iloc[c, r]) and pd.isna(actual.iloc[c, r]):
                pass
            elif isinstance(expected.iloc[c, r], SupportsFloat) and isinstance(actual.iloc[c, r], SupportsFloat):
                if not math.isclose(expected.iloc[c, r], actual.iloc[c, r], rel_tol=1e-6):
                    return False
            elif (
                isinstance(expected.iloc[c, r], pd.Timestamp)
                and isinstance(actual.iloc[c, r], pd.Timestamp)
                # If expected has no tz but actual is UTC, remove timezone. Some engines add UTC by default.
                and actual.iloc[c, r].tzname() == "UTC"
                and expected.iloc[c, r].tzname() is None
            ):
                if actual.iloc[c, r].tz_localize(None) != expected.iloc[c, r].tz_localize(None):
                    return False
            elif expected.iloc[c, r] != actual.iloc[c, r]:
                return False
    return True


def assert_dataframes_equal(
    actual: pd.DataFrame,
    expected: pd.DataFrame,
    sort_columns: bool = True,
    allow_empty: bool = False,
    compare_names_using_lowercase: bool = False,
) -> None:
    """Check that contents of DataFrames are the same.

    If sort_columns is set to false, value and column order needs to be the same.
    If compare_names_using_lowercase is set to True, we copy the dataframes and lower-case their names.
    This is useful for Snowflake query output comparisons.
    """
    if compare_names_using_lowercase:
        actual = actual.copy()
        expected = expected.copy()
        actual.columns = actual.columns.str.lower()
        expected.columns = expected.columns.str.lower()

    if set(actual.columns) != set(expected.columns):
        raise ValueError(
            f"DataFrames do not contain the same columns. actual: {set(actual.columns)}, "
            f"expected: {set(expected.columns)}"
        )

    if not allow_empty and actual.shape[0] == 0 and expected.shape[0] == 0:
        raise AssertionError("Both dataframes have no rows; likely there is a mistake with the test")

    if sort_columns:
        sort_by = list(sorted(actual.columns.tolist()))
        expected = expected.loc[:, sort_by].sort_values(sort_by).reset_index(drop=True)
        actual = actual.loc[:, sort_by].sort_values(sort_by).reset_index(drop=True)

    if not _dataframes_contain_same_data(actual=actual, expected=expected):
        raise ValueError(
            f"Dataframes not equal.\n"
            f"Expected:\n{expected.to_markdown(index=False)}"
            "\n---\n"
            f"Actual:\n{actual.to_markdown(index=False)}"
        )



from __future__ import annotations

import logging
from typing import Sequence

import pytest
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.dataflow.builder.dataflow_plan_builder import DataflowPlanBuilder
from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.node_evaluator import (
    JoinLinkableInstancesRecipe,
    LinkableInstanceSatisfiabilityEvaluation,
    NodeEvaluatorForLinkableInstances,
)
from metricflow.dataflow.builder.partitions import PartitionTimeDimensionJoinDescription
from metricflow.dataflow.dataflow_plan import BaseOutput, ValidityWindowJoinDescription
from metricflow.dataset.dataset import DataSet
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.plan_conversion.node_processor import PreJoinNodeProcessor
from metricflow.specs.specs import (
    DimensionSpec,
    EntityReference,
    EntitySpec,
    LinkableInstanceSpec,
    LinklessEntitySpec,
    TimeDimensionSpec,
)
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository

logger = logging.getLogger(__name__)


@pytest.fixture
def node_evaluator(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
    dataflow_plan_builder: DataflowPlanBuilder,
) -> NodeEvaluatorForLinkableInstances:  # noqa: D
    """Return a node evaluator using the nodes in semantic_model_name_to_nodes."""
    node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        semantic_manifest_lookup=simple_semantic_manifest_lookup,
    )

    source_nodes = tuple(consistent_id_object_repository.simple_model_read_nodes.values())

    return NodeEvaluatorForLinkableInstances(
        semantic_model_lookup=simple_semantic_manifest_lookup.semantic_model_lookup,
        # Use all nodes in the simple model as candidates for joins.
        nodes_available_for_joins=source_nodes,
        node_data_set_resolver=node_data_set_resolver,
    )


def make_multihop_node_evaluator(
    model_source_nodes: Sequence[BaseOutput],
    semantic_manifest_lookup_with_multihop_links: SemanticManifestLookup,
    desired_linkable_specs: Sequence[LinkableInstanceSpec],
) -> NodeEvaluatorForLinkableInstances:  # noqa: D
    """Return a node evaluator using the nodes in multihop_semantic_model_name_to_nodes."""
    node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DunderColumnAssociationResolver(semantic_manifest_lookup_with_multihop_links),
        semantic_manifest_lookup=semantic_manifest_lookup_with_multihop_links,
    )

    node_processor = PreJoinNodeProcessor(
        semantic_model_lookup=semantic_manifest_lookup_with_multihop_links.semantic_model_lookup,
        node_data_set_resolver=node_data_set_resolver,
    )

    nodes_available_for_joins = node_processor.remove_unnecessary_nodes(
        desired_linkable_specs=desired_linkable_specs,
        nodes=model_source_nodes,
        metric_time_dimension_reference=DataSet.metric_time_dimension_reference(),
    )

    nodes_available_for_joins = node_processor.add_multi_hop_joins(
        desired_linkable_specs=desired_linkable_specs, nodes=nodes_available_for_joins
    )

    return NodeEvaluatorForLinkableInstances(
        semantic_model_lookup=semantic_manifest_lookup_with_multihop_links.semantic_model_lookup,
        nodes_available_for_joins=nodes_available_for_joins,
        node_data_set_resolver=node_data_set_resolver,
    )


def test_node_evaluator_with_no_linkable_specs(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(required_linkable_specs=[], start_node=bookings_source_node)
    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(), joinable_linkable_specs=(), join_recipes=(), unjoinable_linkable_specs=()
    )


def test_node_evaluator_with_unjoinable_specs(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="verification_type",
                entity_links=(EntityReference(element_name="verification"),),
            )
        ],
        start_node=bookings_source_node,
    )
    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(),
        join_recipes=(),
        unjoinable_linkable_specs=(
            DimensionSpec(
                element_name="verification_type",
                entity_links=(EntityReference(element_name="verification"),),
            ),
        ),
    )


def test_node_evaluator_with_local_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the requested linkable spec in available in the start node."""
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),))],
        start_node=bookings_source_node,
    )
    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),),
        joinable_linkable_specs=(),
        join_recipes=(),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_local_spec_using_primary_entity(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the requested linkable spec with an entity link is available in the start node."""
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["users_latest"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(element_name="home_state_latest", entity_links=(EntityReference(element_name="user"),))
        ],
        start_node=bookings_source_node,
    )

    assert evaluation == (
        LinkableInstanceSatisfiabilityEvaluation(
            local_linkable_specs=(
                DimensionSpec(
                    element_name="home_state_latest",
                    entity_links=(EntityReference(element_name="user"),),
                ),
            ),
            joinable_linkable_specs=(),
            join_recipes=(),
            unjoinable_linkable_specs=(),
        )
    )


def test_node_evaluator_with_joined_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the requested linkable spec is available if another node is joined."""
    bookings_source_node = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),
            DimensionSpec(
                element_name="country_latest",
                entity_links=(EntityReference(element_name="listing"),),
            ),
            DimensionSpec(
                element_name="capacity_latest",
                entity_links=(EntityReference(element_name="listing"),),
            ),
        ],
        start_node=bookings_source_node,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(DimensionSpec(element_name="is_instant", entity_links=(EntityReference("booking"),)),),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="country_latest",
                entity_links=(EntityReference(element_name="listing"),),
            ),
            DimensionSpec(
                element_name="capacity_latest",
                entity_links=(EntityReference(element_name="listing"),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["listings_latest"],
                join_on_entity=LinklessEntitySpec.from_element_name("listing"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="country_latest",
                        entity_links=(EntityReference(element_name="listing"),),
                    ),
                    DimensionSpec(
                        element_name="capacity_latest",
                        entity_links=(EntityReference(element_name="listing"),),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_joined_spec_on_unique_id(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Similar to test_node_evaluator_with_joined_spec() but using a unique entity."""
    listings_node = consistent_id_object_repository.simple_model_read_nodes["listings_latest"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="company_name",
                entity_links=(EntityReference(element_name="user"),),
            ),
        ],
        start_node=listings_node,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="company_name",
                entity_links=(EntityReference(element_name="user"),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["companies"],
                join_on_entity=LinklessEntitySpec.from_element_name("user"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="company_name",
                        entity_links=(EntityReference(element_name="user"),),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_multiple_joined_specs(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where multiple nodes need to be joined to get all linkable specs."""
    views_source = consistent_id_object_repository.simple_model_read_nodes["views_source"]
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="home_state_latest",
                entity_links=(EntityReference(element_name="user"),),
            ),
            EntitySpec(
                element_name="user",
                entity_links=(EntityReference(element_name="listing"),),
            ),
        ],
        start_node=views_source,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            EntitySpec(
                element_name="user",
                entity_links=(EntityReference(element_name="listing"),),
            ),
            DimensionSpec(
                element_name="home_state_latest",
                entity_links=(EntityReference(element_name="user"),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["listings_latest"],
                join_on_entity=LinklessEntitySpec.from_element_name("listing"),
                satisfiable_linkable_specs=[
                    EntitySpec(
                        element_name="user",
                        entity_links=(EntityReference(element_name="listing"),),
                    )
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["users_latest"],
                join_on_entity=LinklessEntitySpec.from_element_name("user"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="home_state_latest",
                        entity_links=(EntityReference(element_name="user"),),
                    )
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_multihop_joined_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests the case where multiple nodes need to be joined to get all linkable specs."""
    txn_source = consistent_id_object_repository.multihop_model_read_nodes["account_month_txns"]

    linkable_specs = [
        DimensionSpec(
            element_name="customer_name",
            entity_links=(
                EntityReference(element_name="account_id"),
                EntityReference(element_name="customer_id"),
            ),
        ),
    ]

    multihop_node_evaluator = make_multihop_node_evaluator(
        model_source_nodes=consistent_id_object_repository.multihop_model_source_nodes,
        semantic_manifest_lookup_with_multihop_links=multi_hop_join_semantic_manifest_lookup,
        desired_linkable_specs=linkable_specs,
    )

    evaluation = multihop_node_evaluator.evaluate_node(
        required_linkable_specs=linkable_specs,
        start_node=txn_source,
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="customer_name",
                entity_links=(
                    EntityReference(element_name="account_id"),
                    EntityReference(element_name="customer_id"),
                ),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=evaluation.join_recipes[0].node_to_join,
                join_on_entity=LinklessEntitySpec.from_element_name("account_id"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="customer_name",
                        entity_links=(
                            EntityReference(element_name="account_id"),
                            EntityReference(element_name="customer_id"),
                        ),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(
                    PartitionTimeDimensionJoinDescription(
                        start_node_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned", entity_links=(), time_granularity=TimeGranularity.DAY
                        ),
                        node_to_join_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned",
                            entity_links=(),
                            time_granularity=TimeGranularity.DAY,
                        ),
                    ),
                ),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_partition_joined_spec(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    node_evaluator: NodeEvaluatorForLinkableInstances,
) -> None:
    """Tests the case where the joined node required a partitioned join."""
    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="home_state",
                entity_links=(EntityReference(element_name="user"),),
            ),
        ],
        start_node=consistent_id_object_repository.simple_model_read_nodes["id_verifications"],
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="home_state",
                entity_links=(EntityReference(element_name="user"),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.simple_model_read_nodes["users_ds_source"],
                join_on_entity=LinklessEntitySpec.from_element_name("user"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="home_state",
                        entity_links=(EntityReference(element_name="user"),),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(
                    PartitionTimeDimensionJoinDescription(
                        start_node_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned",
                            entity_links=(),
                        ),
                        node_to_join_time_dimension_spec=TimeDimensionSpec(
                            element_name="ds_partitioned",
                            entity_links=(),
                        ),
                    ),
                ),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_scd_target(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    scd_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests the case where the joined node is an SCD with a validity window filter."""
    node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver = DataflowPlanNodeOutputDataSetResolver(
        column_association_resolver=DunderColumnAssociationResolver(scd_semantic_manifest_lookup),
        semantic_manifest_lookup=scd_semantic_manifest_lookup,
    )

    source_nodes = tuple(consistent_id_object_repository.scd_model_read_nodes.values())

    node_evaluator = NodeEvaluatorForLinkableInstances(
        semantic_model_lookup=scd_semantic_manifest_lookup.semantic_model_lookup,
        # Use all nodes in the simple model as candidates for joins.
        nodes_available_for_joins=source_nodes,
        node_data_set_resolver=node_data_set_resolver,
    )

    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=[
            DimensionSpec(
                element_name="is_lux",
                entity_links=(EntityReference(element_name="listing"),),
            )
        ],
        start_node=consistent_id_object_repository.scd_model_read_nodes["bookings_source"],
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="is_lux",
                entity_links=(EntityReference(element_name="listing"),),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=consistent_id_object_repository.scd_model_read_nodes["listings"],
                join_on_entity=LinklessEntitySpec.from_element_name("listing"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="is_lux",
                        entity_links=(EntityReference(element_name="listing"),),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
                validity_window=ValidityWindowJoinDescription(
                    window_start_dimension=TimeDimensionSpec(element_name="window_start", entity_links=()),
                    window_end_dimension=TimeDimensionSpec(element_name="window_end", entity_links=()),
                ),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_multi_hop_scd_target(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    scd_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests the case where the joined node is an SCD reached through another node.

    The validity window should have an entity link, the validity window join is mediated by an intervening
    node, and so we need to refer to that column via the link prefix.
    """
    linkable_specs = [DimensionSpec.from_name("listing__lux_listing__is_confirmed_lux")]
    node_evaluator = make_multihop_node_evaluator(
        model_source_nodes=consistent_id_object_repository.scd_model_source_nodes,
        semantic_manifest_lookup_with_multihop_links=scd_semantic_manifest_lookup,
        desired_linkable_specs=linkable_specs,
    )

    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=linkable_specs,
        start_node=consistent_id_object_repository.scd_model_read_nodes["bookings_source"],
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="is_confirmed_lux",
                entity_links=(
                    EntityReference(element_name="listing"),
                    EntityReference(element_name="lux_listing"),
                ),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=evaluation.join_recipes[0].node_to_join,
                join_on_entity=LinklessEntitySpec.from_element_name("listing"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="is_confirmed_lux",
                        entity_links=(
                            EntityReference(element_name="listing"),
                            EntityReference(element_name="lux_listing"),
                        ),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
                validity_window=ValidityWindowJoinDescription(
                    window_start_dimension=TimeDimensionSpec(
                        element_name="window_start", entity_links=(EntityReference(element_name="lux_listing"),)
                    ),
                    window_end_dimension=TimeDimensionSpec(
                        element_name="window_end", entity_links=(EntityReference(element_name="lux_listing"),)
                    ),
                ),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_multi_hop_through_scd(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    scd_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests the case where the joined node is reached via an SCD.

    The validity window should NOT have any entity links, as the validity window join is not mediated by an
    intervening node and therefore the column name does not use the link prefix.
    """
    linkable_specs = [DimensionSpec.from_name("listing__user__home_state_latest")]
    node_evaluator = make_multihop_node_evaluator(
        model_source_nodes=consistent_id_object_repository.scd_model_source_nodes,
        semantic_manifest_lookup_with_multihop_links=scd_semantic_manifest_lookup,
        desired_linkable_specs=linkable_specs,
    )

    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=linkable_specs,
        start_node=consistent_id_object_repository.scd_model_read_nodes["bookings_source"],
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(
            DimensionSpec(
                element_name="home_state_latest",
                entity_links=(
                    EntityReference(element_name="listing"),
                    EntityReference(element_name="user"),
                ),
            ),
        ),
        join_recipes=(
            JoinLinkableInstancesRecipe(
                node_to_join=evaluation.join_recipes[0].node_to_join,
                join_on_entity=LinklessEntitySpec.from_element_name("listing"),
                satisfiable_linkable_specs=[
                    DimensionSpec(
                        element_name="home_state_latest",
                        entity_links=(
                            EntityReference(element_name="listing"),
                            EntityReference(element_name="user"),
                        ),
                    ),
                ],
                join_on_partition_dimensions=(),
                join_on_partition_time_dimensions=(),
                validity_window=ValidityWindowJoinDescription(
                    window_start_dimension=TimeDimensionSpec(element_name="window_start", entity_links=()),
                    window_end_dimension=TimeDimensionSpec(element_name="window_end", entity_links=()),
                ),
            ),
        ),
        unjoinable_linkable_specs=(),
    )


def test_node_evaluator_with_invalid_multi_hop_scd(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    scd_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests the case where the joined node is reached via an illegal SCD <-> SCD join.

    This will return an empty result because the linkable spec is not joinable in this model.
    """
    linkable_specs = [DimensionSpec.from_name("listing__user__account_type")]
    node_evaluator = make_multihop_node_evaluator(
        model_source_nodes=consistent_id_object_repository.scd_model_source_nodes,
        semantic_manifest_lookup_with_multihop_links=scd_semantic_manifest_lookup,
        desired_linkable_specs=linkable_specs,
    )

    evaluation = node_evaluator.evaluate_node(
        required_linkable_specs=linkable_specs,
        start_node=consistent_id_object_repository.scd_model_read_nodes["bookings_source"],
    )

    assert evaluation == LinkableInstanceSatisfiabilityEvaluation(
        local_linkable_specs=(),
        joinable_linkable_specs=(),
        join_recipes=(),
        unjoinable_linkable_specs=(
            DimensionSpec(
                element_name="account_type",
                entity_links=(
                    EntityReference(element_name="listing"),
                    EntityReference(element_name="user"),
                ),
            ),
        ),
    )



from __future__ import annotations

from metricflow.instances import InstanceSet
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.column_resolver import DunderColumnAssociationResolver
from metricflow.plan_conversion.instance_converters import (
    CreateSelectColumnsWithMeasuresAggregated,
    FilterElements,
)
from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.specs.specs import InstanceSpecSet, MeasureSpec, MetricInputMeasureSpec
from metricflow.sql.sql_exprs import (
    SqlAggregateFunctionExpression,
    SqlFunction,
    SqlPercentileExpression,
)
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository

__SOURCE_TABLE_ALIAS = "a"


def __get_filtered_measure_instance_set(
    semantic_model_name: str, measure_name: str, object_repo: ConsistentIdObjectRepository
) -> InstanceSet:
    """Gets an InstanceSet consisting of only the measure instance matching the given name and semantic model."""
    dataset = object_repo.simple_model_data_sets[semantic_model_name]
    instance_set = dataset.instance_set
    include_specs = tuple(
        instance.spec for instance in instance_set.measure_instances if instance.spec.element_name == measure_name
    )
    return FilterElements(include_specs=InstanceSpecSet(measure_specs=include_specs)).transform(instance_set)


def test_sum_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for booking_value, a SUM type metric in the simple model."""
    measure_name = "booking_value"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="booking_value")),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlAggregateFunctionExpression)
    assert expr.sql_function == SqlFunction.SUM


def test_sum_boolean_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for instant_bookings, a SUM_BOOLEAN type metric in the simple model."""
    measure_name = "instant_bookings"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="instant_bookings")),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlAggregateFunctionExpression)
    # The SUM_BOOLEAN aggregation type is transformed to SUM at model parsing time
    assert expr.sql_function == SqlFunction.SUM


def test_avg_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for average_booking_value, an AVG type metric in the simple model."""
    measure_name = "average_booking_value"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="average_booking_value")),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlAggregateFunctionExpression)
    assert expr.sql_function == SqlFunction.AVERAGE


def test_count_distinct_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for bookers, a COUNT_DISTINCT type metric in the simple model."""
    measure_name = "bookers"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="bookers")),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlAggregateFunctionExpression)
    assert expr.sql_function == SqlFunction.COUNT_DISTINCT


def test_max_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for largest_listing, a MAX type metric in the simple model."""
    measure_name = "largest_listing"
    instance_set = __get_filtered_measure_instance_set("listings_latest", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="largest_listing")),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlAggregateFunctionExpression)
    assert expr.sql_function == SqlFunction.MAX


def test_min_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for smallest_listing, a MIN type metric in the simple model."""
    measure_name = "smallest_listing"
    instance_set = __get_filtered_measure_instance_set("listings_latest", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="smallest_listing")),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlAggregateFunctionExpression)
    assert expr.sql_function == SqlFunction.MIN


def test_aliased_sum(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for booking_value, a SUM type metric in the simple model, with an alias."""
    measure_name = "booking_value"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="booking_value"), alias="bvalue"),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlAggregateFunctionExpression)
    assert expr.sql_function == SqlFunction.SUM
    assert measure_column.column_alias == "bvalue"


def test_percentile_aggregation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Checks for function expression handling for booking_value, a percentile type metric in the simple model."""
    measure_name = "booking_value_p99"
    instance_set = __get_filtered_measure_instance_set("bookings_source", measure_name, consistent_id_object_repository)

    select_column_set: SelectColumnSet = CreateSelectColumnsWithMeasuresAggregated(
        __SOURCE_TABLE_ALIAS,
        DunderColumnAssociationResolver(simple_semantic_manifest_lookup),
        simple_semantic_manifest_lookup.semantic_model_lookup,
        (MetricInputMeasureSpec(measure_spec=MeasureSpec(element_name="booking_value_p99")),),
    ).transform(instance_set=instance_set)

    assert len(select_column_set.measure_columns) == 1
    measure_column = select_column_set.measure_columns[0]
    expr = measure_column.expr
    assert isinstance(expr, SqlPercentileExpression)
    assert expr.percentile_args.percentile == 0.99



from __future__ import annotations

from typing import Dict, Sequence

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.protocols.entity import EntityType
from dbt_semantic_interfaces.references import EntityReference, SemanticModelReference

from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.model.semantics.semantic_model_join_evaluator import (
    SemanticModelEntityJoin,
    SemanticModelEntityJoinType,
    SemanticModelJoinEvaluator,
    SemanticModelLink,
)
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository


def _get_join_types_for_entity_type(entity_type: EntityType) -> Sequence[SemanticModelEntityJoinType]:
    """Exhaustively evaluate entity types and return a sequence of all possible join type pairs.

    The exhaustive conditional statically enforces that every entity type is handled on the left.
    The complete set of matching join types ensures that all pairs are used.
    """
    if (
        entity_type is EntityType.FOREIGN
        or entity_type is EntityType.PRIMARY
        or entity_type is EntityType.UNIQUE
        or entity_type is EntityType.NATURAL
    ):
        join_types = tuple(
            SemanticModelEntityJoinType(left_entity_type=entity_type, right_entity_type=join_type)
            for join_type in EntityType
        )
        return join_types
    else:
        assert_values_exhausted(entity_type)


def test_join_type_coverage() -> None:
    """Ensures all entity type pairs are handled somewhere in the valid/invalid join mapping sets.

    This will prevent surprise RuntimeErrors in production by raising static exceptions for unhandled entity types
    and triggering a test failure for types which are handled in a non-exhaustive fashion
    """
    all_join_types = set(
        SemanticModelJoinEvaluator._INVALID_ENTITY_JOINS + SemanticModelJoinEvaluator._VALID_ENTITY_JOINS
    )
    for entity_type in EntityType:
        join_types = _get_join_types_for_entity_type(entity_type=entity_type)
        for join_type in join_types:
            assert (
                join_type in all_join_types
            ), f"Unhandled entity join type {join_type} not in valid or invalid entity join lists!"


def __get_simple_model_user_semantic_model_references_by_type(
    semantic_manifest_lookup: SemanticManifestLookup,
) -> Dict[EntityType, SemanticModelReference]:
    """Helper to get a set of semantic models with the `user` identifier organized by identifier type."""
    foreign_user_semantic_model = semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("listings_latest")
    )
    primary_user_semantic_model = semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("users_latest")
    )
    unique_user_semantic_model = semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("companies")
    )

    assert foreign_user_semantic_model, "Could not find `listings_latest` semantic model in simple model!"
    assert primary_user_semantic_model, "Could not find `users_latest` semantic model in simple model!"
    assert unique_user_semantic_model, "Could not find `companies` semantic model in simple model!"

    return {
        EntityType.FOREIGN: foreign_user_semantic_model.reference,
        EntityType.PRIMARY: primary_user_semantic_model.reference,
        EntityType.UNIQUE: unique_user_semantic_model.reference,
    }


def test_distinct_target_semantic_model_join_validation(
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests semantic model join validation to a PRIMARY or UNIQUE entity.

    PRIMARY and UNIQUE entity targets should be valid for any join at the semantic model level because they both
    represent entity columns with distinct value sets, and as such there is no risk of inadvertent fanout joins.
    """
    semantic_model_references = __get_simple_model_user_semantic_model_references_by_type(
        simple_semantic_manifest_lookup
    )
    user_entity_reference = EntityReference(element_name="user")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=simple_semantic_manifest_lookup.semantic_model_lookup
    )

    foreign_primary = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.FOREIGN],
        right_semantic_model_reference=semantic_model_references[EntityType.PRIMARY],
        on_entity_reference=user_entity_reference,
    )
    primary_primary = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.PRIMARY],
        right_semantic_model_reference=semantic_model_references[EntityType.PRIMARY],
        on_entity_reference=user_entity_reference,
    )
    unique_primary = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.UNIQUE],
        right_semantic_model_reference=semantic_model_references[EntityType.PRIMARY],
        on_entity_reference=user_entity_reference,
    )
    foreign_unique = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.FOREIGN],
        right_semantic_model_reference=semantic_model_references[EntityType.UNIQUE],
        on_entity_reference=user_entity_reference,
    )
    primary_unique = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.PRIMARY],
        right_semantic_model_reference=semantic_model_references[EntityType.UNIQUE],
        on_entity_reference=user_entity_reference,
    )
    unique_unique = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.UNIQUE],
        right_semantic_model_reference=semantic_model_references[EntityType.UNIQUE],
        on_entity_reference=user_entity_reference,
    )

    results = {
        "foreign to primary": foreign_primary,
        "primary to primary": primary_primary,
        "unique to primary": unique_primary,
        "foreign to unique": foreign_unique,
        "primary to unique": primary_unique,
        "unique to unique": unique_unique,
    }
    assert all(results.values()), (
        f"All semantic model level join types for primary and unique targets should be valid, but we found "
        f"at least one that was not! Incorrectly failing types: {[k for k,v in results.items() if not v]}."
    )


def test_foreign_target_semantic_model_join_validation(simple_semantic_manifest_lookup: SemanticManifestLookup) -> None:
    """Tests semantic model join validation to FOREIGN entity types.

    These should all fail by default, as fanout joins are not supported
    """
    semantic_model_references = __get_simple_model_user_semantic_model_references_by_type(
        simple_semantic_manifest_lookup
    )
    user_entity_reference = EntityReference(element_name="user")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=simple_semantic_manifest_lookup.semantic_model_lookup
    )

    foreign_foreign = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.FOREIGN],
        right_semantic_model_reference=semantic_model_references[EntityType.FOREIGN],
        on_entity_reference=user_entity_reference,
    )
    primary_foreign = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.PRIMARY],
        right_semantic_model_reference=semantic_model_references[EntityType.FOREIGN],
        on_entity_reference=user_entity_reference,
    )
    unique_foreign = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=semantic_model_references[EntityType.UNIQUE],
        right_semantic_model_reference=semantic_model_references[EntityType.FOREIGN],
        on_entity_reference=user_entity_reference,
    )

    results = {
        "foreign to foreign": foreign_foreign,
        "primary to foreign": primary_foreign,
        "unique to foreign": unique_foreign,
    }
    assert not any(results.values()), (
        f"All semantic model level joins against foreign targets should be invalid, but we found at least one "
        f"that was not! Incorrectly passing types: {[k for k,v in results.items() if v]}."
    )


def test_semantic_model_join_validation_on_missing_entity(
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests semantic model join validation where the entity is missing from one or both semantic models."""
    primary_listing_semantic_model = simple_semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("listings_latest")
    )
    assert primary_listing_semantic_model, "Could not find semantic model `listings_latest` in the simple model!"
    no_listing_semantic_model = simple_semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("id_verifications")
    )
    assert no_listing_semantic_model, "Could not find semantic model `id_verifications` in the simple model!"
    listing_entity_reference = EntityReference(element_name="listing")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=simple_semantic_manifest_lookup.semantic_model_lookup
    )

    assert not join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=no_listing_semantic_model.reference,
        right_semantic_model_reference=primary_listing_semantic_model.reference,
        on_entity_reference=listing_entity_reference,
    ), (
        "Found valid join on `listing` involving the `id_verifications` semantic model, which does not include the "
        "`listing` entity!"
    )


def test_distinct_target_instance_set_join_validation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests instance set join validation to a PRIMARY or UNIQUE entity."""
    foreign_user_instance_set = consistent_id_object_repository.simple_model_data_sets["listings_latest"].instance_set
    primary_user_instance_set = consistent_id_object_repository.simple_model_data_sets["users_latest"].instance_set
    unique_user_instance_set = consistent_id_object_repository.simple_model_data_sets["companies"].instance_set
    user_entity_reference = EntityReference(element_name="user")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=simple_semantic_manifest_lookup.semantic_model_lookup
    )

    foreign_primary = join_evaluator.is_valid_instance_set_join(
        left_instance_set=foreign_user_instance_set,
        right_instance_set=primary_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    primary_primary = join_evaluator.is_valid_instance_set_join(
        left_instance_set=primary_user_instance_set,
        right_instance_set=primary_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    unique_primary = join_evaluator.is_valid_instance_set_join(
        left_instance_set=unique_user_instance_set,
        right_instance_set=primary_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    foreign_unique = join_evaluator.is_valid_instance_set_join(
        left_instance_set=foreign_user_instance_set,
        right_instance_set=unique_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    primary_unique = join_evaluator.is_valid_instance_set_join(
        left_instance_set=primary_user_instance_set,
        right_instance_set=unique_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    unique_unique = join_evaluator.is_valid_instance_set_join(
        left_instance_set=unique_user_instance_set,
        right_instance_set=unique_user_instance_set,
        on_entity_reference=user_entity_reference,
    )

    results = {
        "foreign to primary": foreign_primary,
        "primary to primary": primary_primary,
        "unique to primary": unique_primary,
        "foreign to unique": foreign_unique,
        "primary to unique": primary_unique,
        "unique to unique": unique_unique,
    }
    assert all(results.values()), (
        f"All instance set level join types for primary and unique targets should be valid, but we found "
        f"at least one that was not! Incorrectly failing types: {[k for k,v in results.items() if not v]}."
    )


def test_foreign_target_instance_set_join_validation(
    consistent_id_object_repository: ConsistentIdObjectRepository,
    simple_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    """Tests semantic model join validation to FOREIGN entity types."""
    foreign_user_instance_set = consistent_id_object_repository.simple_model_data_sets["listings_latest"].instance_set
    primary_user_instance_set = consistent_id_object_repository.simple_model_data_sets["users_latest"].instance_set
    unique_user_instance_set = consistent_id_object_repository.simple_model_data_sets["companies"].instance_set
    user_entity_reference = EntityReference(element_name="user")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=simple_semantic_manifest_lookup.semantic_model_lookup
    )

    foreign_foreign = join_evaluator.is_valid_instance_set_join(
        left_instance_set=foreign_user_instance_set,
        right_instance_set=foreign_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    primary_foreign = join_evaluator.is_valid_instance_set_join(
        left_instance_set=primary_user_instance_set,
        right_instance_set=foreign_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    unique_foreign = join_evaluator.is_valid_instance_set_join(
        left_instance_set=unique_user_instance_set,
        right_instance_set=foreign_user_instance_set,
        on_entity_reference=user_entity_reference,
    )

    results = {
        "foreign to foreign": foreign_foreign,
        "primary to foreign": primary_foreign,
        "unique to foreign": unique_foreign,
    }
    assert not any(results.values()), (
        f"All semantic model level joins against foreign targets should be invalid, but we found at least one "
        f"that was not! Incorrectly passing types: {[k for k,v in results.items() if v]}."
    )


def test_get_joinable_semantic_models_single_hop(  # noqa: D
    multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    semantic_model_reference = SemanticModelReference(semantic_model_name="account_month_txns")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=multi_hop_join_semantic_manifest_lookup.semantic_model_lookup
    )

    # Single-hop
    joinable_semantic_models = join_evaluator.get_joinable_semantic_models(
        left_semantic_model_reference=semantic_model_reference
    )
    assert set(joinable_semantic_models.keys()) == {"bridge_table"}
    assert joinable_semantic_models["bridge_table"] == SemanticModelLink(
        left_semantic_model_reference=SemanticModelReference(semantic_model_name="account_month_txns"),
        join_path=[
            SemanticModelEntityJoin(
                right_semantic_model_reference=SemanticModelReference(semantic_model_name="bridge_table"),
                entity_reference=EntityReference(element_name="account_id"),
                join_type=SemanticModelEntityJoinType(
                    left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.UNIQUE
                ),
            )
        ],
    )


def test_get_joinable_semantic_models_multi_hop(  # noqa: D
    multi_hop_join_semantic_manifest_lookup: SemanticManifestLookup,
) -> None:
    semantic_model_reference = SemanticModelReference(semantic_model_name="account_month_txns")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=multi_hop_join_semantic_manifest_lookup.semantic_model_lookup
    )

    # 2-hop
    joinable_semantic_models = join_evaluator.get_joinable_semantic_models(
        left_semantic_model_reference=semantic_model_reference, include_multi_hop=True
    )
    assert set(joinable_semantic_models.keys()) == {"bridge_table", "customer_other_data", "customer_table"}
    assert joinable_semantic_models["bridge_table"] == SemanticModelLink(
        left_semantic_model_reference=SemanticModelReference(semantic_model_name="account_month_txns"),
        join_path=[
            SemanticModelEntityJoin(
                right_semantic_model_reference=SemanticModelReference(semantic_model_name="bridge_table"),
                entity_reference=EntityReference(element_name="account_id"),
                join_type=SemanticModelEntityJoinType(
                    left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.UNIQUE
                ),
            )
        ],
    )
    assert joinable_semantic_models["customer_other_data"] == SemanticModelLink(
        left_semantic_model_reference=SemanticModelReference(semantic_model_name="account_month_txns"),
        join_path=[
            SemanticModelEntityJoin(
                right_semantic_model_reference=SemanticModelReference(semantic_model_name="bridge_table"),
                entity_reference=EntityReference(element_name="account_id"),
                join_type=SemanticModelEntityJoinType(
                    left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.UNIQUE
                ),
            ),
            SemanticModelEntityJoin(
                right_semantic_model_reference=SemanticModelReference(semantic_model_name="customer_other_data"),
                entity_reference=EntityReference(element_name="customer_id"),
                join_type=SemanticModelEntityJoinType(
                    left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.PRIMARY
                ),
            ),
        ],
    )
    assert joinable_semantic_models["customer_table"] == SemanticModelLink(
        left_semantic_model_reference=SemanticModelReference(semantic_model_name="account_month_txns"),
        join_path=[
            SemanticModelEntityJoin(
                right_semantic_model_reference=SemanticModelReference(semantic_model_name="bridge_table"),
                entity_reference=EntityReference(element_name="account_id"),
                join_type=SemanticModelEntityJoinType(
                    left_entity_type=EntityType.PRIMARY, right_entity_type=EntityType.UNIQUE
                ),
            ),
            SemanticModelEntityJoin(
                right_semantic_model_reference=SemanticModelReference(semantic_model_name="customer_table"),
                entity_reference=EntityReference(element_name="customer_id"),
                join_type=SemanticModelEntityJoinType(
                    left_entity_type=EntityType.FOREIGN, right_entity_type=EntityType.PRIMARY
                ),
            ),
        ],
    )


def test_natural_entity_semantic_model_validation(scd_semantic_manifest_lookup: SemanticManifestLookup) -> None:
    """Tests semantic model validation for NATURAL target entity types.

    These tests rely on the scd_semantic_manifest_lookup, which makes extensive use of NATURAL key types.
    """
    natural_user_semantic_model = scd_semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("primary_accounts")
    )
    primary_user_semantic_model = scd_semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("users_latest")
    )
    foreign_user_semantic_model = scd_semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("bookings_source")
    )
    unique_user_semantic_model = scd_semantic_manifest_lookup.semantic_model_lookup.get_by_reference(
        SemanticModelReference("companies")
    )
    user_entity_reference = EntityReference(element_name="user")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=scd_semantic_manifest_lookup.semantic_model_lookup
    )
    # Type refinement
    assert natural_user_semantic_model, "Could not find `primary_accounts` semantic model in scd model!"
    assert foreign_user_semantic_model, "Could not find `bookings_source` semantic model in scd model!"
    assert primary_user_semantic_model, "Could not find `users_latest` semantic model in scd model!"
    assert unique_user_semantic_model, "Could not find `companies` semantic model in scd model!"

    # Valid cases
    natural_primary = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=natural_user_semantic_model.reference,
        right_semantic_model_reference=primary_user_semantic_model.reference,
        on_entity_reference=user_entity_reference,
    )
    natural_unique = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=natural_user_semantic_model.reference,
        right_semantic_model_reference=unique_user_semantic_model.reference,
        on_entity_reference=user_entity_reference,
    )
    foreign_natural = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=foreign_user_semantic_model.reference,
        right_semantic_model_reference=natural_user_semantic_model.reference,
        on_entity_reference=user_entity_reference,
    )
    primary_natural = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=primary_user_semantic_model.reference,
        right_semantic_model_reference=natural_user_semantic_model.reference,
        on_entity_reference=user_entity_reference,
    )
    unique_natural = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=unique_user_semantic_model.reference,
        right_semantic_model_reference=natural_user_semantic_model.reference,
        on_entity_reference=user_entity_reference,
    )
    # Invalid cases
    natural_foreign = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=natural_user_semantic_model.reference,
        right_semantic_model_reference=foreign_user_semantic_model.reference,
        on_entity_reference=user_entity_reference,
    )
    natural_natural = join_evaluator.is_valid_semantic_model_join(
        left_semantic_model_reference=natural_user_semantic_model.reference,
        right_semantic_model_reference=natural_user_semantic_model.reference,
        on_entity_reference=user_entity_reference,
    )

    valid_joins = {
        "natural to primary": natural_primary,
        "natural to unique": natural_unique,
        "foreign to natural": foreign_natural,
        "primary to natural": primary_natural,
        "unique to natural": unique_natural,
    }
    invalid_joins = {
        "natural to foreign": natural_foreign,
        "natural to natural": natural_natural,
    }
    assert all(valid_joins.values()) and not any(invalid_joins.values()), (
        f"Found unexpected join validator results when validating joins involving natural key comparisons! Valid "
        f"joins marked invalid: {[k for k,v in valid_joins.items() if not v]}. Invalid joins marked valid: "
        f"{[k for k, v in invalid_joins.items() if v]}."
    )


def test_natural_entity_instance_set_validation(
    consistent_id_object_repository: ConsistentIdObjectRepository, scd_semantic_manifest_lookup: SemanticManifestLookup
) -> None:
    """Tests instance set validation for NATURAL target entity types.

    These tests rely on the scd_semantic_manifest_lookup, which makes extensive use of NATURAL key types.
    """
    natural_user_instance_set = consistent_id_object_repository.scd_model_data_sets["primary_accounts"].instance_set
    primary_user_instance_set = consistent_id_object_repository.scd_model_data_sets["users_latest"].instance_set
    foreign_user_instance_set = consistent_id_object_repository.scd_model_data_sets["bookings_source"].instance_set
    unique_user_instance_set = consistent_id_object_repository.scd_model_data_sets["companies"].instance_set
    user_entity_reference = EntityReference(element_name="user")
    join_evaluator = SemanticModelJoinEvaluator(
        semantic_model_lookup=scd_semantic_manifest_lookup.semantic_model_lookup
    )

    # Valid cases
    natural_primary = join_evaluator.is_valid_instance_set_join(
        left_instance_set=natural_user_instance_set,
        right_instance_set=primary_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    natural_unique = join_evaluator.is_valid_instance_set_join(
        left_instance_set=natural_user_instance_set,
        right_instance_set=unique_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    foreign_natural = join_evaluator.is_valid_instance_set_join(
        left_instance_set=foreign_user_instance_set,
        right_instance_set=natural_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    primary_natural = join_evaluator.is_valid_instance_set_join(
        left_instance_set=primary_user_instance_set,
        right_instance_set=natural_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    unique_natural = join_evaluator.is_valid_instance_set_join(
        left_instance_set=unique_user_instance_set,
        right_instance_set=natural_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    # Invalid cases
    natural_foreign = join_evaluator.is_valid_instance_set_join(
        left_instance_set=natural_user_instance_set,
        right_instance_set=foreign_user_instance_set,
        on_entity_reference=user_entity_reference,
    )
    natural_natural = join_evaluator.is_valid_instance_set_join(
        left_instance_set=natural_user_instance_set,
        right_instance_set=natural_user_instance_set,
        on_entity_reference=user_entity_reference,
    )

    valid_joins = {
        "natural to primary": natural_primary,
        "natural to unique": natural_unique,
        "foreign to natural": foreign_natural,
        "primary to natural": primary_natural,
        "unique to natural": unique_natural,
    }
    invalid_joins = {
        "natural to foreign": natural_foreign,
        "natural to natural": natural_natural,
    }
    assert all(valid_joins.values()) and not any(invalid_joins.values()), (
        f"Found unexpected join validator results when validating joins involving natural key comparisons! Valid "
        f"joins marked invalid: {[k for k,v in valid_joins.items() if not v]}. Invalid joins marked valid: "
        f"{[k for k, v in invalid_joins.items() if v]}."
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

T = TypeVar("T")


class SemanticModelContainer(ABC, Generic[T]):  # noqa: D
    @abstractmethod
    def get(self, semantic_model_name: str) -> T:  # noqa: D
        pass

    @abstractmethod
    def values(self) -> List[T]:  # noqa: D
        pass

    @abstractmethod
    def keys(self) -> List[str]:  # noqa: D
        pass

    @abstractmethod
    def __contains__(self, item: str) -> bool:  # noqa: D
        pass

    @abstractmethod
    def put(self, key: str, value: T) -> None:  # noqa: D
        pass



from __future__ import annotations

import pytest
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity

from metricflow.dataflow.dataflow_plan import ValidityWindowJoinDescription
from metricflow.instances import InstanceSet
from metricflow.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow.plan_conversion.instance_converters import CreateValidityWindowJoinDescription
from metricflow.specs.specs import TimeDimensionSpec
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository


def test_no_validity_dims(
    consistent_id_object_repository: ConsistentIdObjectRepository, scd_semantic_manifest_lookup: SemanticManifestLookup
) -> None:
    """Tests converting an instance set with no matching dimensions to a ValidityWindowJoinDescription."""
    # bookings_source is a fact table, and has no validity window dimensions
    dataset = consistent_id_object_repository.scd_model_data_sets["bookings_source"]

    validity_window_join_description = CreateValidityWindowJoinDescription(
        semantic_model_lookup=scd_semantic_manifest_lookup.semantic_model_lookup
    ).transform(instance_set=dataset.instance_set)

    assert validity_window_join_description is None, (
        f"We managed to create a validity window join description `{validity_window_join_description}` from a "
        f"semantic model that does not have one defined!"
    )


def test_validity_window_conversion(
    consistent_id_object_repository: ConsistentIdObjectRepository, scd_semantic_manifest_lookup: SemanticManifestLookup
) -> None:
    """Tests converting an instance set with a single validity window into a ValidityWindowJoinDescription."""
    # The listings semantic model uses a 2-column SCD Type III layout
    dataset = consistent_id_object_repository.scd_model_data_sets["listings"]
    expected_join_description = ValidityWindowJoinDescription(
        window_start_dimension=TimeDimensionSpec(
            element_name="window_start",
            time_granularity=TimeGranularity.DAY,
            entity_links=(),
        ),
        window_end_dimension=TimeDimensionSpec(
            element_name="window_end", time_granularity=TimeGranularity.DAY, entity_links=()
        ),
    )

    validity_window_join_description = CreateValidityWindowJoinDescription(
        semantic_model_lookup=scd_semantic_manifest_lookup.semantic_model_lookup
    ).transform(instance_set=dataset.instance_set)

    assert (
        validity_window_join_description is not None
    ), "Failed to make a validity window join description from a dataset which should have one configured!"
    assert (
        validity_window_join_description == expected_join_description
    ), f"Expected validity window: `{expected_join_description}` but got: `{validity_window_join_description}`"


def test_multiple_validity_windows(
    consistent_id_object_repository: ConsistentIdObjectRepository, scd_semantic_manifest_lookup: SemanticManifestLookup
) -> None:
    """Tests the behavior of this converter when it encounters an instance set with multiple validity windows."""
    first_dataset = consistent_id_object_repository.scd_model_data_sets["listings"]
    second_dataset = consistent_id_object_repository.scd_model_data_sets["primary_accounts"]
    merged_instance_set = InstanceSet.merge([first_dataset.instance_set, second_dataset.instance_set])
    with pytest.raises(AssertionError, match="Found more than 1 set of validity window specs"):
        CreateValidityWindowJoinDescription(
            semantic_model_lookup=scd_semantic_manifest_lookup.semantic_model_lookup
        ).transform(merged_instance_set)



from __future__ import annotations

from _pytest.fixtures import FixtureRequest

from metricflow.dag.id_generation import OPTIMIZED_DATAFLOW_PLAN_PREFIX, IdGeneratorRegistry
from metricflow.dataflow.dataflow_plan import (
    BaseOutput,
    DataflowPlan,
    FilterElementsNode,
    WriteToResultDataframeNode,
)
from metricflow.dataflow.dataflow_plan_to_text import dataflow_plan_as_text
from metricflow.dataflow.optimizer.source_scan.cm_branch_combiner import (
    ComputeMetricsBranchCombiner,
    ComputeMetricsBranchCombinerResult,
)
from metricflow.specs.specs import InstanceSpecSet, MeasureSpec
from metricflow.test.dataflow_plan_to_svg import display_graph_if_requested
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.snapshot_utils import assert_plan_snapshot_text_equal


def make_dataflow_plan(node: BaseOutput) -> DataflowPlan:  # noqa: D
    return DataflowPlan(
        plan_id=IdGeneratorRegistry.for_class(ComputeMetricsBranchCombiner).create_id(OPTIMIZED_DATAFLOW_PLAN_PREFIX),
        sink_output_nodes=[WriteToResultDataframeNode(node)],
    )


def test_read_sql_source_combination(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    """Tests combining a single node."""
    source0 = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    source1 = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    combiner = ComputeMetricsBranchCombiner(source0)

    result: ComputeMetricsBranchCombinerResult = source1.accept(combiner)
    assert result.combined_branch

    dataflow_plan = make_dataflow_plan(result.combined_branch)
    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=dataflow_plan,
        plan_snapshot_text=dataflow_plan_as_text(dataflow_plan),
    )

    display_graph_if_requested(
        request=request,
        mf_test_session_state=mf_test_session_state,
        dag_graph=dataflow_plan,
    )


def test_filter_combination(  # noqa: D
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    consistent_id_object_repository: ConsistentIdObjectRepository,
) -> None:
    """Tests combining a single node."""
    source0 = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    filter0 = FilterElementsNode(
        parent_node=source0, include_specs=InstanceSpecSet(measure_specs=(MeasureSpec(element_name="bookings"),))
    )
    source1 = consistent_id_object_repository.simple_model_read_nodes["bookings_source"]
    filter1 = FilterElementsNode(
        parent_node=source1,
        include_specs=InstanceSpecSet(
            measure_specs=(MeasureSpec(element_name="booking_value"),),
        ),
    )
    combiner = ComputeMetricsBranchCombiner(filter0)

    result: ComputeMetricsBranchCombinerResult = filter1.accept(combiner)
    assert result.combined_branch

    dataflow_plan = make_dataflow_plan(result.combined_branch)
    assert_plan_snapshot_text_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        plan=dataflow_plan,
        plan_snapshot_text=dataflow_plan_as_text(dataflow_plan),
    )

    display_graph_if_requested(
        request=request,
        mf_test_session_state=mf_test_session_state,
        dag_graph=dataflow_plan,
    )
